# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Adapts the grpc surface."""

from __future__ import absolute_import

from grpc import RpcError, StatusCode

from google.gax import _grpc_google_auth


API_ERRORS = (RpcError, )
"""gRPC exceptions that indicate that an RPC was aborted."""


STATUS_CODE_NAMES = {
    'ABORTED': StatusCode.ABORTED,
    'CANCELLED': StatusCode.CANCELLED,
    'DATA_LOSS': StatusCode.DATA_LOSS,
    'DEADLINE_EXCEEDED': StatusCode.DEADLINE_EXCEEDED,
    'FAILED_PRECONDITION': StatusCode.FAILED_PRECONDITION,
    'INTERNAL': StatusCode.INTERNAL,
    'INVALID_ARGUMENT': StatusCode.INVALID_ARGUMENT,
    'NOT_FOUND': StatusCode.NOT_FOUND,
    'OUT_OF_RANGE': StatusCode.OUT_OF_RANGE,
    'PERMISSION_DENIED': StatusCode.PERMISSION_DENIED,
    'RESOURCE_EXHAUSTED': StatusCode.RESOURCE_EXHAUSTED,
    'UNAUTHENTICATED': StatusCode.UNAUTHENTICATED,
    'UNAVAILABLE': StatusCode.UNAVAILABLE,
    'UNIMPLEMENTED': StatusCode.UNIMPLEMENTED,
    'UNKNOWN': StatusCode.UNKNOWN}
"""Maps strings used in client config to gRPC status codes."""


NAME_STATUS_CODES = dict([(v, k) for (k, v) in STATUS_CODE_NAMES.items()])
"""Inverse map for STATUS_CODE_NAMES"""


def exc_to_code(exc):
    """Retrieves the status code from an exception"""
    if not isinstance(exc, RpcError):
        return None
    else:
        try:
            return exc.code()
        except AttributeError:
            return None


def create_stub(generated_create_stub, channel=None, service_path=None,
                service_port=None, credentials=None, scopes=None,
                ssl_credentials=None):
    """Creates a gRPC client stub.

    Args:
        generated_create_stub (Callable): The generated gRPC method to create a
            stub.
        channel (grpc.Channel): A Channel object through which to make calls.
            If None, a secure channel is constructed. If specified, all
            remaining arguments are ignored.
        service_path (str): The domain name of the API remote host.
        service_port (int): The port on which to connect to the remote host.
        credentials (google.auth.credentials.Credentials): The authorization
            credentials to attach to requests. These credentials identify your
            application to the service.
        scopes (Sequence[str]): The OAuth scopes for this service. This
            parameter is ignored if a credentials is specified.
        ssl_credentials (grpc.ChannelCredentials): gRPC channel credentials
            used to create a secure gRPC channel. If not specified, SSL
            credentials will be created using default certificates.

    Returns:
        grpc.Client: A gRPC client stub.
    """
    if channel is None:
        target = '{}:{}'.format(service_path, service_port)

        if credentials is None:
            credentials = _grpc_google_auth.get_default_credentials(scopes)

        channel = _grpc_google_auth.secure_authorized_channel(
            credentials, target, ssl_credentials=ssl_credentials)

    return generated_create_stub(channel)



# Copyright 2016, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Provides behavior that supports request bundling.

:func:`compute_bundle_id` is used generate ids linking API requests to the
appropriate bundles.

:class:`Event` is the result of scheduling a bundled api call.  It is a
decorated :class:`threading.Event`; its ``wait`` and ``is_set`` methods
are used wait for the bundle request to complete or determine if it has
been completed respectively.

:class:`Task` manages the sending of all the requests in a specific bundle.

:class:`Executor` has a ``schedule`` method that is used add bundled api calls
to a new or existing :class:`Task`.

"""

from __future__ import absolute_import

import collections
import copy
import logging
import threading

_LOG = logging.getLogger(__name__)


def _str_dotted_getattr(obj, name):
    """Expands extends getattr to allow dots in x to indicate nested objects.

    Args:
       obj (object): an object.
       name (str): a name for a field in the object.

    Returns:
       Any: the value of named attribute.

    Raises:
       AttributeError: if the named attribute does not exist.
    """
    for part in name.split('.'):
        obj = getattr(obj, part)
    return str(obj) if obj else None


def compute_bundle_id(obj, discriminator_fields):
    """Computes a bundle id from the discriminator fields of `obj`.

    discriminator_fields may include '.' as a separator, which is used to
    indicate object traversal.  This is meant to allow fields in the
    computed bundle_id.

    the id is a tuple computed by going through the discriminator fields in
    order and obtaining the str(value) object field (or nested object field)

    if any discriminator field cannot be found, ValueError is raised.

    Args:
      obj (object): an object.
      discriminator_fields (Sequence[str]): a list of discriminator fields in
        the order to be to be used in the id.

    Returns:
      Tuple[str]: computed as described above.

    Raises:
      AttributeError: if any discriminator fields attribute does not exist.
    """
    return tuple(_str_dotted_getattr(obj, x) for x in discriminator_fields)


_WARN_DEMUX_MISMATCH = ('cannot demultiplex the bundled response, got'
                        ' %d subresponses; want %d, each bundled request will'
                        ' receive all responses')


class Task(object):
    """Coordinates the execution of a single bundle."""
    # pylint: disable=too-many-instance-attributes

    def __init__(self, api_call, bundle_id, bundled_field, bundling_request,
                 kwargs, subresponse_field=None):
        """
        Args:
           api_call (Callable[Sequence[object], object]): the func that is this
             tasks's API call.
           bundle_id (Tuple[str]): the id of this bundle.
           bundled_field (str): the field used to create the bundled request.
           bundling_request (object): the request to pass as the arg to
              api_call.
           kwargs (dict): keyword arguments passed to api_call.
           subresponse_field (str): optional field used to demultiplex
              responses.

        """
        self._api_call = api_call
        self._bundling_request = bundling_request
        self._kwargs = kwargs
        self.bundle_id = bundle_id
        self.bundled_field = bundled_field
        self.subresponse_field = subresponse_field
        self.timer = None
        self._in_deque = collections.deque()
        self._event_deque = collections.deque()

    @property
    def element_count(self):
        """The number of bundled elements in the repeated field."""
        return sum(len(elts) for elts in self._in_deque)

    @property
    def request_bytesize(self):
        """The size of in bytes of the bundled field elements."""
        return sum(len(str(e)) for elts in self._in_deque for e in elts)

    def run(self):
        """Call the task's func.

        The task's func will be called with the bundling requests func
        """
        if len(self._in_deque) == 0:
            return
        req = self._bundling_request
        del getattr(req, self.bundled_field)[:]
        getattr(req, self.bundled_field).extend(
            [e for elts in self._in_deque for e in elts])

        subresponse_field = self.subresponse_field
        if subresponse_field:
            self._run_with_subresponses(req, subresponse_field, self._kwargs)
        else:
            self._run_with_no_subresponse(req, self._kwargs)

    def _run_with_no_subresponse(self, req, kwargs):
        try:
            resp = self._api_call(req, **kwargs)
            for event in self._event_deque:
                event.result = resp
                event.set()
        except Exception as exc:  # pylint: disable=broad-except
            for event in self._event_deque:
                event.result = exc
                event.set()
        finally:
            self._in_deque.clear()
            self._event_deque.clear()

    def _run_with_subresponses(self, req, subresponse_field, kwargs):
        try:
            resp = self._api_call(req, **kwargs)
            in_sizes = [len(elts) for elts in self._in_deque]
            all_subresponses = getattr(resp, subresponse_field)
            if len(all_subresponses) != sum(in_sizes):
                _LOG.warning(_WARN_DEMUX_MISMATCH, len(all_subresponses),
                             sum(in_sizes))
                for event in self._event_deque:
                    event.result = resp
                    event.set()
            else:
                start = 0
                for i, event in zip(in_sizes, self._event_deque):
                    next_copy = copy.copy(resp)
                    subresponses = all_subresponses[start:start + i]
                    next_copy.ClearField(subresponse_field)
                    getattr(next_copy, subresponse_field).extend(subresponses)
                    start += i
                    event.result = next_copy
                    event.set()
        except Exception as exc:  # pylint: disable=broad-except
            for event in self._event_deque:
                event.result = exc
                event.set()
        finally:
            self._in_deque.clear()
            self._event_deque.clear()

    def extend(self, elts):
        """Adds elts to the tasks.

        Args:
           elts (Sequence): a iterable of elements that can be appended to the
            task's bundle_field.

        Returns:
            Event: an event that can be used to wait on the response.
        """
        # Use a copy, not a reference, as it is later necessary to mutate
        # the proto field from which elts are drawn in order to construct
        # the bundled request.
        elts = elts[:]
        self._in_deque.append(elts)
        event = self._event_for(elts)
        self._event_deque.append(event)
        return event

    def _event_for(self, elts):
        """Creates an Event that is set when the bundle with elts is sent."""
        event = Event()
        event.canceller = self._canceller_for(elts, event)
        return event

    def _canceller_for(self, elts, event):
        """Obtains a cancellation function that removes elts.

        The returned cancellation function returns ``True`` if all elements
        was removed successfully from the _in_deque, and false if it was not.
        """
        def canceller():
            """Cancels submission of ``elts`` as part of this bundle.

            Returns:
               bool: ``False`` if any of elements had already been sent,
               otherwise ``True``.
            """
            try:
                self._event_deque.remove(event)
                self._in_deque.remove(elts)
                return True
            except ValueError:
                return False

        return canceller


TIMER_FACTORY = threading.Timer  # pylint: disable=invalid-name
"""A class with an interface similar to threading.Timer.

Defaults to threading.Timer.  This makes it easy to plug-in alternate
timer implementations."""


class Executor(object):
    """Organizes bundling for an api service that requires it."""
    # pylint: disable=too-few-public-methods

    def __init__(self, options):
        """Constructor.

        Args:
           options (gax.BundleOptions): configures strategy this instance
             uses when executing bundled functions.

        """
        self._options = options
        self._tasks = {}
        self._task_lock = threading.RLock()

    def schedule(self, api_call, bundle_id, bundle_desc, bundling_request,
                 kwargs=None):
        """Schedules bundle_desc of bundling_request as part of bundle_id.

        The returned value an :class:`Event` that

        * has a ``result`` attribute that will eventually be set to the result
          the api call
        * will be used to wait for the response
        * holds the canceller function for canceling this part of the bundle

        Args:
          api_call (callable[[object], object]): the scheduled API call.
          bundle_id (str): identifies the bundle on which the API call should be
            made.
          bundle_desc (gax.BundleDescriptor): describes the structure of the
            bundled call.
          bundling_request (object): the request instance to use in the API
            call.
          kwargs (dict): optional, the keyword arguments passed to the API call.

        Returns:
           Event: the scheduled event.
        """
        kwargs = kwargs or dict()
        bundle = self._bundle_for(api_call, bundle_id, bundle_desc,
                                  bundling_request, kwargs)
        elts = getattr(bundling_request, bundle_desc.bundled_field)
        event = bundle.extend(elts)

        # Run the bundle if the count threshold was reached.
        count_threshold = self._options.element_count_threshold
        if count_threshold > 0 and bundle.element_count >= count_threshold:
            self._run_now(bundle.bundle_id)

        # Run the bundle if the size threshold was reached.
        size_threshold = self._options.request_byte_threshold
        if size_threshold > 0 and bundle.request_bytesize >= size_threshold:
            self._run_now(bundle.bundle_id)

        return event

    def _bundle_for(self, api_call, bundle_id, bundle_desc, bundling_request,
                    kwargs):
        with self._task_lock:
            bundle = self._tasks.get(bundle_id)
            if bundle is None:
                bundle = Task(api_call, bundle_id, bundle_desc.bundled_field,
                              bundling_request, kwargs,
                              subresponse_field=bundle_desc.subresponse_field)
                delay_threshold = self._options.delay_threshold
                if delay_threshold > 0:
                    self._run_later(bundle, delay_threshold)
                self._tasks[bundle_id] = bundle
            return bundle

    def _run_later(self, bundle, delay_threshold):
        with self._task_lock:
            if bundle.timer is None:
                the_timer = TIMER_FACTORY(
                    delay_threshold,
                    self._run_now,
                    args=[bundle.bundle_id])
                the_timer.start()
                bundle.timer = the_timer

    def _run_now(self, bundle_id):
        with self._task_lock:
            if bundle_id in self._tasks:
                a_task = self._tasks.pop(bundle_id)
                a_task.run()


class Event(object):
    """Wraps a threading.Event, adding, canceller and result attributes."""

    def __init__(self):
        """Constructor.

        """
        self._event = threading.Event()
        self.result = None
        self.canceller = None

    def is_set(self):
        """Calls ``is_set`` on the decorated :class:`threading.Event`."""
        return self._event.is_set()

    def set(self):
        """Calls ``set`` on the decorated :class:`threading.Event`."""
        return self._event.set()

    def clear(self):
        """Calls ``clear`` on the decorated :class:`threading.Event`.

        Also resets the result if one has been set.
        """
        self.result = None
        return self._event.clear()

    def wait(self, timeout=None):
        """Calls ``wait`` on the decorated :class:`threading.Event`."""
        return self._event.wait(timeout=timeout)

    def cancel(self):
        """Invokes the cancellation function provided on construction."""
        if self.canceller:
            return self.canceller()
        else:
            return False



# Copyright 2017, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import

import collections
import platform
import unittest

import pkg_resources

from google import gax
from google.gax.utils import metrics


# pylint: disable=no-member
GRPC_VERSION = pkg_resources.get_distribution('grpcio').version
# pylint: enable=no-member


class TestFill(unittest.TestCase):
    def test_no_argument(self):
        headers = metrics.fill()

        # Assert that the headers are set appropriately.
        self.assertEqual(headers['gl-python'], platform.python_version())
        self.assertEqual(headers['gax'], gax.__version__)
        self.assertEqual(headers['grpc'], GRPC_VERSION)

        # Assert that the headers are in the correct order.
        self.assertEqual(
            [k for k in headers.keys()],
            ['gl-python', 'gax', 'grpc'],
        )

    def test_gapic(self):
        headers = metrics.fill(collections.OrderedDict((
            ('gl-python', platform.python_version()),
            ('gapic', '1.0.0'),
        )))

        # Assert that the headers are set appropriately.
        self.assertEqual(headers['gl-python'], platform.python_version())
        self.assertEqual(headers['gapic'], '1.0.0')
        self.assertEqual(headers['gax'], gax.__version__)
        self.assertEqual(headers['grpc'], GRPC_VERSION)

        # Assert that the headers are in the correct order.
        self.assertEqual(
            [k for k in headers.keys()],
            ['gl-python', 'gapic', 'gax', 'grpc'],
        )


class TestStringify(unittest.TestCase):
    def test_no_argument(self):
        self.assertEqual(metrics.stringify(), '')

    def test_with_unordered_argument(self):
        string = metrics.stringify({
            'gl-python': platform.python_version(),
            'gapic': '1.0.0',
            'gax': gax.__version__,
            'grpc': GRPC_VERSION,
        })

        # Assert that each section of the string is present, but
        # ignore ordering.
        self.assertIn('gl-python/%s' % platform.python_version(), string)
        self.assertIn('gapic/1.0.0', string)
        self.assertIn('gax/%s' % gax.__version__, string)
        self.assertIn('grpc/%s' % GRPC_VERSION, string)

    def test_with_ordered_argument(self):
        headers = collections.OrderedDict()
        headers['gl-python'] = platform.python_version()
        headers['gapic'] = '1.0.0'
        headers['gax'] = gax.__version__
        headers['grpc'] = GRPC_VERSION
        string = metrics.stringify(headers)

        # Check for the exact string, order and all.
        expected = 'gl-python/{python} gapic/{gapic} gax/{gax} grpc/{grpc}'
        self.assertEqual(string, expected.format(
            gapic='1.0.0',
            gax=gax.__version__,
            grpc=GRPC_VERSION,
            python=platform.python_version(),
        ))



# Copyright 2017, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Utility functions for manipulation of metrics headers."""

from __future__ import absolute_import

import collections
import platform

import pkg_resources

from google import gax


def fill(metrics_headers=()):
    """Add the metrics headers known to GAX.

    Return an OrderedDict with all of the metrics headers provided to
    this function, as well as the metrics known to GAX (such as its own
    version, the GRPC version, etc.).
    """
    # Create an ordered dictionary with the Python version, which
    # should go first.
    answer = collections.OrderedDict((
        ('gl-python', platform.python_version()),
    ))

    # Add anything that already appears in the passed metrics headers,
    # in order.
    for key, value in collections.OrderedDict(metrics_headers).items():
        answer[key] = value

    # Add the GAX and GRPC headers to our metrics.
    # These come after what may have been passed in (generally the GAPIC
    # library).
    answer['gax'] = gax.__version__
    # pylint: disable=no-member
    answer['grpc'] = pkg_resources.get_distribution('grpcio').version
    # pylint: enable=no-member

    return answer


def stringify(metrics_headers=()):
    """Convert the provided metrics headers to a string.

    Iterate over the metrics headers (a dictionary, usually ordered) and
    return a properly-formatted space-separated string
    (e.g. foo/1.2.3 bar/3.14.159).
    """
    metrics_headers = collections.OrderedDict(metrics_headers)
    return ' '.join(['%s/%s' % (k, v) for k, v in metrics_headers.items()])



# Copyright 2016, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pylint: disable=missing-docstring,no-self-use,no-init,invalid-name,protected-access,too-many-public-methods
"""Unit tests for gax package globals."""

from __future__ import absolute_import

import logging
import multiprocessing as mp

import mock
import unittest2

from google.gax import (
    _CallSettings, _LOG, _OperationFuture, BundleOptions, CallOptions,
    INITIAL_PAGE, OPTION_INHERIT, RetryOptions)
from google.gax.errors import GaxError, RetryError
from google.longrunning import operations_pb2
from google.rpc import code_pb2, status_pb2

from tests.fixtures.fixture_pb2 import Simple


class TestBundleOptions(unittest2.TestCase):

    def test_cannot_construct_with_noarg_options(self):
        self.assertRaises(AssertionError,
                          BundleOptions)

    def test_cannot_construct_with_bad_options(self):
        not_an_int = 'i am a string'
        self.assertRaises(AssertionError,
                          BundleOptions,
                          element_count_threshold=not_an_int)
        self.assertRaises(AssertionError,
                          BundleOptions,
                          request_byte_threshold=not_an_int)
        self.assertRaises(AssertionError,
                          BundleOptions,
                          delay_threshold=not_an_int)


class TestCallSettings(unittest2.TestCase):

    def test_call_options_simple(self):
        options = CallOptions(timeout=23)
        self.assertEqual(options.timeout, 23)
        self.assertEqual(options.retry, OPTION_INHERIT)
        self.assertEqual(options.page_token, OPTION_INHERIT)

    def test_cannot_construct_bad_options(self):
        self.assertRaises(
            ValueError, CallOptions, timeout=47, retry=RetryOptions(None, None))

    def test_settings_merge_options1(self):
        options = CallOptions(timeout=46)
        settings = _CallSettings(timeout=9, page_descriptor=None, retry=None)
        final = settings.merge(options)
        self.assertEqual(final.timeout, 46)
        self.assertIsNone(final.retry)
        self.assertIsNone(final.page_descriptor)

    def test_settings_merge_options2(self):
        retry = RetryOptions(None, None)
        options = CallOptions(retry=retry)
        settings = _CallSettings(
            timeout=9, page_descriptor=None, retry=RetryOptions(None, None))
        final = settings.merge(options)
        self.assertEqual(final.timeout, 9)
        self.assertIsNone(final.page_descriptor)
        self.assertEqual(final.retry, retry)

    def test_settings_merge_options_page_streaming(self):
        retry = RetryOptions(None, None)
        page_descriptor = object()
        options = CallOptions(timeout=46, page_token=INITIAL_PAGE)
        settings = _CallSettings(timeout=9, retry=retry,
                                 page_descriptor=page_descriptor)
        final = settings.merge(options)
        self.assertEqual(final.timeout, 46)
        self.assertEqual(final.page_descriptor, page_descriptor)
        self.assertEqual(final.page_token, INITIAL_PAGE)
        self.assertFalse(final.flatten_pages)
        self.assertEqual(final.retry, retry)

    def test_settings_merge_none(self):
        settings = _CallSettings(
            timeout=23, page_descriptor=object(), bundler=object(),
            retry=object())
        final = settings.merge(None)
        self.assertEqual(final.timeout, settings.timeout)
        self.assertEqual(final.retry, settings.retry)
        self.assertEqual(final.page_descriptor, settings.page_descriptor)
        self.assertEqual(final.bundler, settings.bundler)
        self.assertEqual(final.bundle_descriptor, settings.bundle_descriptor)


def _task1(operation_future):
    operation_future.test_queue.put(operation_future.result().field1)


def _task2(operation_future):
    operation_future.test_queue.put(operation_future.result().field2)


class _FakeOperationsClient(object):
    def __init__(self, operations):
        self.operations = list(reversed(operations))

    def get_operation(self, *_):
        return self.operations.pop()

    def cancel_operation(self, *_):
        pass


class _FakeLoggingHandler(logging.Handler):
    def __init__(self, *args, **kwargs):
        self.queue = mp.Queue()
        super(_FakeLoggingHandler, self).__init__(*args, **kwargs)

    def emit(self, record):
        self.acquire()
        try:
            self.queue.put(record.getMessage())
        finally:
            self.release()

    def reset(self):
        self.acquire()
        try:
            self.queue = mp.Queue()
        finally:
            self.release()


class TestOperationFuture(unittest2.TestCase):

    OPERATION_NAME = 'operations/projects/foo/instances/bar/operations/123'

    @classmethod
    def setUpClass(cls):
        cls._log_handler = _FakeLoggingHandler(level='DEBUG')
        _LOG.addHandler(cls._log_handler)

    def setUp(self):
        self._log_handler.reset()

    def _make_operation(self, metadata=None, response=None, error=None,
                        **kwargs):
        operation = operations_pb2.Operation(name=self.OPERATION_NAME, **kwargs)

        if metadata is not None:
            operation.metadata.Pack(metadata)

        if response is not None:
            operation.response.Pack(response)

        if error is not None:
            operation.error.CopyFrom(error)

        return operation

    def _make_operation_future(self, *operations):
        if not operations:
            operations = [self._make_operation()]

        fake_client = _FakeOperationsClient(operations)
        return _OperationFuture(operations[0], fake_client, Simple, Simple)

    def test_cancel_issues_call_when_not_done(self):
        operation = self._make_operation()

        fake_client = _FakeOperationsClient([operation])
        fake_client.cancel_operation = mock.Mock()

        operation_future = _OperationFuture(
            operation, fake_client, Simple, Simple)

        self.assertTrue(operation_future.cancel())
        fake_client.cancel_operation.assert_called_with(self.OPERATION_NAME)

    def test_cancel_does_nothing_when_already_done(self):
        operation = self._make_operation(done=True)

        fake_client = _FakeOperationsClient([operation])
        fake_client.cancel_operation = mock.Mock()

        operation_future = _OperationFuture(
            operation, fake_client, Simple, Simple)

        self.assertFalse(operation_future.cancel())
        fake_client.cancel_operation.assert_not_called()

    def test_cancelled_true(self):
        error = status_pb2.Status(code=code_pb2.CANCELLED)
        operation = self._make_operation(error=error)
        operation_future = self._make_operation_future(operation)

        self.assertTrue(operation_future.cancelled())

    def test_cancelled_false(self):
        operation = self._make_operation(error=status_pb2.Status())
        operation_future = self._make_operation_future(operation)
        self.assertFalse(operation_future.cancelled())

    def test_done_true(self):
        operation = self._make_operation(done=True)
        operation_future = self._make_operation_future(operation)
        self.assertTrue(operation_future.done())

    def test_done_false(self):
        operation_future = self._make_operation_future()
        self.assertFalse(operation_future.done())

    def test_operation_name(self):
        operation_future = self._make_operation_future()
        self.assertEqual(self.OPERATION_NAME, operation_future.operation_name())

    def test_metadata(self):
        metadata = Simple()
        operation = self._make_operation(metadata=metadata)
        operation_future = self._make_operation_future(operation)

        self.assertEqual(metadata, operation_future.metadata())

    def test_metadata_none(self):
        operation_future = self._make_operation_future()
        self.assertIsNone(operation_future.metadata())

    def test_last_operation_data(self):
        operation = self._make_operation()
        operation_future = self._make_operation_future(operation)
        self.assertEqual(operation, operation_future.last_operation_data())

    def test_result_response(self):
        response = Simple()
        operation = self._make_operation(done=True, response=response)
        operation_future = self._make_operation_future(operation)

        self.assertEqual(response, operation_future.result())

    def test_result_error(self):
        operation = self._make_operation(done=True, error=status_pb2.Status())
        operation_future = self._make_operation_future(operation)
        self.assertRaises(GaxError, operation_future.result)

    def test_result_timeout(self):
        operation_future = self._make_operation_future()
        self.assertRaises(RetryError, operation_future.result, 0)

    def test_exception_error(self):
        error = status_pb2.Status()
        operation = self._make_operation(done=True, error=error)
        operation_future = self._make_operation_future(operation)

        self.assertEqual(error, operation_future.exception())

    def test_exception_response(self):
        operation = self._make_operation(done=True, response=Simple())
        operation_future = self._make_operation_future(operation)
        self.assertIsNone(operation_future.exception())

    def test_exception_timeout(self):
        operation_future = self._make_operation_future()
        self.assertRaises(RetryError, operation_future.exception, 0)

    def test_add_done_callback(self):
        response = Simple(field1='foo', field2='bar')
        operation_future = self._make_operation_future(
            self._make_operation(),
            self._make_operation(done=True, response=response))
        operation_future.test_queue = mp.Queue()

        operation_future.add_done_callback(_task1)
        operation_future.add_done_callback(_task2)

        self.assertEqual('foo', operation_future.test_queue.get())
        self.assertEqual('bar', operation_future.test_queue.get())

    def test_add_done_callback_when_already_done(self):
        response = Simple(field1='foo', field2='bar')
        operation_future = self._make_operation_future(
            self._make_operation(done=True, response=response))
        operation_future.test_queue = mp.Queue()

        operation_future.add_done_callback(_task1)

        self.assertEqual('foo', operation_future.test_queue.get())

    def test_add_done_callback_when_exception(self):
        def _raising_task(_):
            raise Exception('Test message')

        operation_future = self._make_operation_future(
            self._make_operation(),
            self._make_operation(done=True, response=Simple()))
        operation_future.add_done_callback(_raising_task)
        self.assertEqual('Test message', self._log_handler.queue.get())



# Copyright 2016, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Provides function wrappers that implement page streaming and retrying."""

from __future__ import absolute_import, division, unicode_literals

from future import utils

from google import gax
from google.gax import bundling
from google.gax.utils import metrics

_MILLIS_PER_SECOND = 1000


def _bundleable(desc):
    """Creates a function that transforms an API call into a bundling call.

    It transform a_func from an API call that receives the requests and returns
    the response into a callable that receives the same request, and
    returns a :class:`bundling.Event`.

    The returned Event object can be used to obtain the eventual result of the
    bundled call.

    Args:
      desc (gax.BundleDescriptor): describes the bundling that a_func
        supports.

    Returns:
      Callable: takes the API call's request and keyword args and returns a
        bundling.Event object.

    """
    def inner(a_func, settings, request, **kwargs):
        """Schedules execution of a bundling task."""
        if not settings.bundler:
            return a_func(request, **kwargs)

        the_id = bundling.compute_bundle_id(
            request, desc.request_discriminator_fields)
        return settings.bundler.schedule(a_func, the_id, desc, request, kwargs)

    return inner


def _page_streamable(page_descriptor):
    """Creates a function that yields an iterable to performs page-streaming.

    Args:
        page_descriptor (:class:`PageDescriptor`): indicates the structure
          of page streaming to be performed.

    Returns:
        Callable: A function that returns an iterator.
    """

    def inner(a_func, settings, request, **kwargs):
        """Actual page-streaming based on the settings."""
        page_iterator = gax.PageIterator(
            a_func, page_descriptor, settings.page_token, request, **kwargs)
        if settings.flatten_pages:
            return gax.ResourceIterator(page_iterator)
        else:
            return page_iterator

    return inner


def _construct_bundling(bundle_config, bundle_descriptor):
    """Helper for ``construct_settings()``.

    Args:
      bundle_config (dict): A dictionary specifying a bundle parameters, the
        value for 'bundling' field in a method config (See
        ``construct_settings()`` for information on this config.)
      bundle_descriptor (BundleDescriptor): A BundleDescriptor object
        describing the structure of bundling for this method. If not set,
        this method will not bundle.

    Returns:
      Tuple[bundling.Executor, BundleDescriptor]: A tuple that configures
        bundling. The bundling.Executor may be None if this method should not
        bundle.
    """
    if bundle_config and bundle_descriptor:
        bundler = bundling.Executor(gax.BundleOptions(
            element_count_threshold=bundle_config.get(
                'element_count_threshold', 0),
            element_count_limit=bundle_config.get('element_count_limit', 0),
            request_byte_threshold=bundle_config.get(
                'request_byte_threshold', 0),
            request_byte_limit=bundle_config.get('request_byte_limit', 0),
            delay_threshold=bundle_config.get('delay_threshold_millis', 0)))
    else:
        bundler = None

    return bundler


def _construct_retry(method_config, retry_codes, retry_params, retry_names):
    """Helper for ``construct_settings()``.

    Args:
      method_config (dict): A dictionary representing a single ``methods``
        entry of the standard API client config file. (See
        ``construct_settings()`` for information on this yaml.)
      retry_codes (dict): A dictionary parsed from the ``retry_codes`` entry
        of the standard API client config file. (See ``construct_settings()``
        for information on this yaml.)
      retry_params (dict): A dictionary parsed from the ``retry_params`` entry
        of the standard API client config file. (See ``construct_settings()``
        for information on this yaml.)
      retry_names (dict): A dictionary mapping the string names used in the
        standard API client config file to API response status codes.

    Returns:
      Optional[RetryOptions]: The retry options, if applicable.
    """
    if method_config is None:
        return None

    codes = None
    if retry_codes and 'retry_codes_name' in method_config:
        codes_name = method_config['retry_codes_name']
        if codes_name in retry_codes and retry_codes[codes_name]:
            codes = [retry_names[name] for name in retry_codes[codes_name]]
        else:
            codes = []

    backoff_settings = None
    if retry_params and 'retry_params_name' in method_config:
        params_name = method_config['retry_params_name']
        if params_name and params_name in retry_params:
            backoff_settings = gax.BackoffSettings(**retry_params[params_name])

    return gax.RetryOptions(
        backoff_settings=backoff_settings,
        retry_codes=codes,
    )


def _merge_retry_options(retry_options, overrides):
    """Helper for ``construct_settings()``.

    Takes two retry options, and merges them into a single RetryOption instance.

    Args:
      retry_options (RetryOptions): The base RetryOptions.
      overrides (RetryOptions): The RetryOptions used for overriding ``retry``.
        Use the values if it is not None. If entire ``overrides`` is None,
        ignore the base retry and return None.

    Returns:
      RetryOptions: The merged options, or None if it will be canceled.
    """
    if overrides is None:
        return None

    if overrides.retry_codes is None and overrides.backoff_settings is None:
        return retry_options

    codes = retry_options.retry_codes
    if overrides.retry_codes is not None:
        codes = overrides.retry_codes
    backoff_settings = retry_options.backoff_settings
    if overrides.backoff_settings is not None:
        backoff_settings = overrides.backoff_settings

    return gax.RetryOptions(
        backoff_settings=backoff_settings,
        retry_codes=codes,
    )


def _upper_camel_to_lower_under(string):
    if not string:
        return ''
    out = ''
    out += string[0].lower()
    for char in string[1:]:
        if char.isupper():
            out += '_' + char.lower()
        else:
            out += char
    return out


def construct_settings(
        service_name, client_config, config_override,
        retry_names, bundle_descriptors=None, page_descriptors=None,
        metrics_headers=(), kwargs=None):
    """Constructs a dictionary mapping method names to _CallSettings.

    The ``client_config`` parameter is parsed from a client configuration JSON
    file of the form:

    .. code-block:: json

       {
         "interfaces": {
           "google.fake.v1.ServiceName": {
             "retry_codes": {
               "idempotent": ["UNAVAILABLE", "DEADLINE_EXCEEDED"],
               "non_idempotent": []
             },
             "retry_params": {
               "default": {
                 "initial_retry_delay_millis": 100,
                 "retry_delay_multiplier": 1.2,
                 "max_retry_delay_millis": 1000,
                 "initial_rpc_timeout_millis": 2000,
                 "rpc_timeout_multiplier": 1.5,
                 "max_rpc_timeout_millis": 30000,
                 "total_timeout_millis": 45000
               }
             },
             "methods": {
               "CreateFoo": {
                 "retry_codes_name": "idempotent",
                 "retry_params_name": "default",
                 "timeout_millis": 30000
               },
               "Publish": {
                 "retry_codes_name": "non_idempotent",
                 "retry_params_name": "default",
                 "bundling": {
                   "element_count_threshold": 40,
                   "element_count_limit": 200,
                   "request_byte_threshold": 90000,
                   "request_byte_limit": 100000,
                   "delay_threshold_millis": 100
                 }
               }
             }
           }
         }
       }

    Args:
      service_name (str): The fully-qualified name of this service, used as a
        key into the client config file (in the example above, this value
        would be ``google.fake.v1.ServiceName``).
      client_config (dict): A dictionary parsed from the standard API client
        config file.
      bundle_descriptors (Mapping[str, BundleDescriptor]): A dictionary of
        method names to BundleDescriptor objects for methods that are
        bundling-enabled.
      page_descriptors (Mapping[str, PageDescriptor]): A dictionary of method
        names to PageDescriptor objects for methods that are page
        streaming-enabled.
      config_override (str): A dictionary in the same structure of
        client_config to override the settings. Usually client_config is
        supplied from the default config and config_override will be
        specified by users.
      retry_names (Mapping[str, object]): A dictionary mapping the strings
        referring to response status codes to the Python objects representing
        those codes.
      metrics_headers (Mapping[str, str]): Dictionary of headers to be passed
        for analytics. Sent as a dictionary; eventually becomes a
        space-separated string (e.g. 'foo/1.0.0 bar/3.14.1').
      kwargs (dict): The keyword arguments to be passed to the API calls.

    Returns:
      dict: A dictionary mapping method names to _CallSettings.

    Raises:
      KeyError: If the configuration for the service in question cannot be
        located in the provided ``client_config``.
    """
    # pylint: disable=too-many-locals
    # pylint: disable=protected-access
    defaults = {}
    bundle_descriptors = bundle_descriptors or {}
    page_descriptors = page_descriptors or {}
    kwargs = kwargs or {}

    # Sanity check: It is possible that we got this far but some headers
    # were specified with an older library, which sends them as...
    #   kwargs={'metadata': [('x-goog-api-client', 'foo/1.0 bar/3.0')]}
    #
    # Note: This is the final format we will send down to GRPC shortly.
    #
    # Remove any x-goog-api-client header that may have been present
    # in the metadata list.
    if 'metadata' in kwargs:
        kwargs['metadata'] = [value for value in kwargs['metadata']
                              if value[0].lower() != 'x-goog-api-client']

    # Fill out the metrics headers with GAX and GRPC info, and convert
    # to a string in the format that the GRPC layer expects.
    kwargs.setdefault('metadata', [])
    kwargs['metadata'].append(
        ('x-goog-api-client', metrics.stringify(metrics.fill(metrics_headers)))
    )

    try:
        service_config = client_config['interfaces'][service_name]
    except KeyError:
        raise KeyError('Client configuration not found for service: {}'
                       .format(service_name))

    overrides = config_override.get('interfaces', {}).get(service_name, {})

    for method in service_config.get('methods'):
        method_config = service_config['methods'][method]
        overriding_method = overrides.get('methods', {}).get(method, {})
        snake_name = _upper_camel_to_lower_under(method)

        if overriding_method and overriding_method.get('timeout_millis'):
            timeout = overriding_method['timeout_millis']
        else:
            timeout = method_config['timeout_millis']
        timeout /= _MILLIS_PER_SECOND

        bundle_descriptor = bundle_descriptors.get(snake_name)
        bundling_config = method_config.get('bundling', None)
        if overriding_method and 'bundling' in overriding_method:
            bundling_config = overriding_method['bundling']
        bundler = _construct_bundling(bundling_config, bundle_descriptor)

        retry_options = _merge_retry_options(
            _construct_retry(method_config, service_config['retry_codes'],
                             service_config['retry_params'], retry_names),
            _construct_retry(overriding_method, overrides.get('retry_codes'),
                             overrides.get('retry_params'), retry_names))

        defaults[snake_name] = gax._CallSettings(
            timeout=timeout, retry=retry_options,
            page_descriptor=page_descriptors.get(snake_name),
            bundler=bundler, bundle_descriptor=bundle_descriptor,
            kwargs=kwargs)
    return defaults


def _catch_errors(a_func, to_catch):
    """Updates a_func to wrap exceptions with GaxError

    Args:
        a_func (callable): A callable.
        to_catch (list[Exception]): Configures the exceptions to wrap.

    Returns:
        Callable: A function that will wrap certain exceptions with GaxError
    """
    def inner(*args, **kwargs):
        """Wraps specified exceptions"""
        try:
            return a_func(*args, **kwargs)
        # pylint: disable=catching-non-exception
        except tuple(to_catch) as exception:
            utils.raise_with_traceback(
                gax.errors.create_error('RPC failed', cause=exception))

    return inner


def create_api_call(func, settings):
    """Converts an rpc call into an API call governed by the settings.

    In typical usage, ``func`` will be a callable used to make an rpc request.
    This will mostly likely be a bound method from a request stub used to make
    an rpc call.

    The result is created by applying a series of function decorators defined
    in this module to ``func``.  ``settings`` is used to determine which
    function decorators to apply.

    The result is another callable which for most values of ``settings`` has
    has the same signature as the original. Only when ``settings`` configures
    bundling does the signature change.

    Args:
      func (Callable[Sequence[object], object]): is used to make a bare rpc
        call.
      settings (_CallSettings): provides the settings for this call

    Returns:
      Callable[Sequence[object], object]: a bound method on a request stub used
        to make an rpc call

    Raises:
       ValueError: if ``settings`` has incompatible values, e.g, if bundling
         and page_streaming are both configured

    """
    def base_caller(api_call, _, *args):
        """Simply call api_call and ignore settings."""
        return api_call(*args)

    def inner(request, options=None):
        """Invoke with the actual settings."""
        this_settings = settings.merge(options)
        if this_settings.retry and this_settings.retry.retry_codes:
            api_call = gax.retry.retryable(
                func, this_settings.retry, **this_settings.kwargs)
        else:
            api_call = gax.retry.add_timeout_arg(
                func, this_settings.timeout, **this_settings.kwargs)
        api_call = _catch_errors(api_call, gax.config.API_ERRORS)
        return api_caller(api_call, this_settings, request)

    if settings.page_descriptor:
        if settings.bundler and settings.bundle_descriptor:
            raise ValueError('The API call has incompatible settings: '
                             'bundling and page streaming')
        api_caller = _page_streamable(settings.page_descriptor)
    elif settings.bundler and settings.bundle_descriptor:
        api_caller = _bundleable(settings.bundle_descriptor)
    else:
        api_caller = base_caller

    return inner



# Copyright 2016, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Implements a utility for parsing and formatting path templates."""

from __future__ import absolute_import

from collections import namedtuple

from ply import lex, yacc

_BINDING = 1
_END_BINDING = 2
_TERMINAL = 3
_Segment = namedtuple('_Segment', ['kind', 'literal'])


def _format(segments):
    template = ''
    slash = True
    for segment in segments:
        if segment.kind == _TERMINAL:
            if slash:
                template += '/'
            template += segment.literal
        slash = True
        if segment.kind == _BINDING:
            template += '/{%s=' % segment.literal
            slash = False
        if segment.kind == _END_BINDING:
            template += '%s}' % segment.literal
    return template[1:]  # Remove the leading /


class ValidationException(Exception):
    """Represents a path template validation error."""
    pass


class PathTemplate(object):
    """Represents a path template."""

    segments = None
    segment_count = 0

    def __init__(self, data):
        parser = _Parser()
        self.segments = parser.parse(data)
        self.segment_count = parser.segment_count

    def __len__(self):
        return self.segment_count

    def __repr__(self):
        return _format(self.segments)

    def render(self, bindings):
        """Renders a string from a path template using the provided bindings.

        Args:
            bindings (dict): A dictionary of var names to binding strings.

        Returns:
            str: The rendered instantiation of this path template.

        Raises:
            ValidationException: If a key isn't provided or if a sub-template
                can't be parsed.
        """
        out = []
        binding = False
        for segment in self.segments:
            if segment.kind == _BINDING:
                if segment.literal not in bindings:
                    raise ValidationException(
                        ('rendering error: value for key \'{}\' '
                         'not provided').format(segment.literal))
                out.extend(PathTemplate(bindings[segment.literal]).segments)
                binding = True
            elif segment.kind == _END_BINDING:
                binding = False
            else:
                if binding:
                    continue
                out.append(segment)
        path = _format(out)
        self.match(path)
        return path

    def match(self, path):
        """Matches a fully qualified path template string.

        Args:
            path (str): A fully qualified path template string.

        Returns:
            dict: Var names to matched binding values.

        Raises:
            ValidationException: If path can't be matched to the template.
        """
        this = self.segments
        that = path.split('/')
        current_var = None
        bindings = {}
        segment_count = self.segment_count
        j = 0
        for i in range(0, len(this)):
            if j >= len(that):
                break
            if this[i].kind == _TERMINAL:
                if this[i].literal == '*':
                    bindings[current_var] = that[j]
                    j += 1
                elif this[i].literal == '**':
                    until = j + len(that) - segment_count + 1
                    segment_count += len(that) - segment_count
                    bindings[current_var] = '/'.join(that[j:until])
                    j = until
                elif this[i].literal != that[j]:
                    raise ValidationException(
                        'mismatched literal: \'%s\' != \'%s\'' % (
                            this[i].literal, that[j]))
                else:
                    j += 1
            elif this[i].kind == _BINDING:
                current_var = this[i].literal
        if j != len(that) or j != segment_count:
            raise ValidationException(
                'match error: could not render from the path template: {}'
                .format(path))
        return bindings


# pylint: disable=C0103
# pylint: disable=R0201
class _Parser(object):
    tokens = (
        'FORWARD_SLASH',
        'LEFT_BRACE',
        'RIGHT_BRACE',
        'EQUALS',
        'WILDCARD',
        'PATH_WILDCARD',
        'LITERAL',
    )

    t_FORWARD_SLASH = r'/'
    t_LEFT_BRACE = r'\{'
    t_RIGHT_BRACE = r'\}'
    t_EQUALS = r'='
    t_WILDCARD = r'\*'
    t_PATH_WILDCARD = r'\*\*'
    t_LITERAL = r'[^*=}{\/]+'

    t_ignore = ' \t'

    binding_var_count = 0
    segment_count = 0

    def __init__(self):
        self.lexer = lex.lex(module=self)
        self.parser = yacc.yacc(module=self, debug=False, write_tables=False)

    def parse(self, data):
        """Returns a list of path template segments parsed from data.

        Args:
            data (str): A path template string.

        Returns:
            Sequence[_PathSegment]: The path template segments.

        Raises:
            ValidationException: If the path template is invalid.
        """
        self.binding_var_count = 0
        self.segment_count = 0

        segments = self.parser.parse(data)
        # Validation step: checks that there are no nested bindings.
        path_wildcard = False
        for segment in segments:
            if segment.kind == _TERMINAL and segment.literal == '**':
                if path_wildcard:
                    raise ValidationException(
                        'validation error: path template cannot contain more '
                        'than one path wildcard')
                path_wildcard = True
        return segments

    def p_template(self, p):
        """template : FORWARD_SLASH bound_segments
                    | bound_segments"""
        # ply fails on a negative index.
        p[0] = p[len(p) - 1]

    def p_bound_segments(self, p):
        """bound_segments : bound_segment FORWARD_SLASH bound_segments
                          | bound_segment"""
        p[0] = p[1]
        if len(p) > 2:
            p[0].extend(p[3])

    def p_unbound_segments(self, p):
        """unbound_segments : unbound_terminal FORWARD_SLASH unbound_segments
                            | unbound_terminal"""
        p[0] = p[1]
        if len(p) > 2:
            p[0].extend(p[3])

    def p_bound_segment(self, p):
        """bound_segment : bound_terminal
                         | variable"""
        p[0] = p[1]

    def p_unbound_terminal(self, p):
        """unbound_terminal : WILDCARD
                            | PATH_WILDCARD
                            | LITERAL"""
        p[0] = [_Segment(_TERMINAL, p[1])]
        self.segment_count += 1

    def p_bound_terminal(self, p):
        """bound_terminal : unbound_terminal"""
        if p[1][0].literal in ['*', '**']:
            p[0] = [_Segment(_BINDING, '$%d' % self.binding_var_count),
                    p[1][0],
                    _Segment(_END_BINDING, '')]
            self.binding_var_count += 1
        else:
            p[0] = p[1]

    def p_variable(self, p):
        """variable : LEFT_BRACE LITERAL EQUALS unbound_segments RIGHT_BRACE
                    | LEFT_BRACE LITERAL RIGHT_BRACE"""
        p[0] = [_Segment(_BINDING, p[2])]
        if len(p) > 4:
            p[0].extend(p[4])
        else:
            p[0].append(_Segment(_TERMINAL, '*'))
            self.segment_count += 1
        p[0].append(_Segment(_END_BINDING, ''))

    def p_error(self, p):
        """Raises a parser error."""
        if p:
            raise ValidationException(
                'parser error: unexpected token \'%s\'' % p.type)
        else:
            raise ValidationException('parser error: unexpected EOF')

    def t_error(self, t):
        """Raises a lexer error."""
        raise ValidationException(
            'lexer error: illegal character \'%s\'' % t.value[0])



# Copyright 2017, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest

import pytest

from google.api import http_pb2
from google.gax.utils import protobuf
from google.longrunning import operations_proto_pb2 as ops


class GetTests(unittest.TestCase):
    def test_get_dict_sentinel(self):
        with pytest.raises(KeyError):
            assert protobuf.get({}, 'foo')

    def test_get_dict_present(self):
        assert protobuf.get({'foo': 'bar'}, 'foo') == 'bar'

    def test_get_dict_default(self):
        assert protobuf.get({}, 'foo', default='bar') == 'bar'

    def test_get_dict_nested(self):
        assert protobuf.get({'foo': {'bar': 'baz'}}, 'foo.bar') == 'baz'

    def test_get_pb2_sentinel(self):
        op = ops.Operation()
        with pytest.raises(KeyError):
            assert protobuf.get(op, 'foo')

    def test_get_pb2_present(self):
        op = ops.Operation(name='foo')
        assert protobuf.get(op, 'name') == 'foo'

    def test_get_pb2_default(self):
        op = ops.Operation()
        assert protobuf.get(op, 'foo', default='bar') == 'bar'

    def test_invalid_object(self):
        obj = object()
        with pytest.raises(TypeError):
            protobuf.get(obj, 'foo', 'bar')


class SetTests(unittest.TestCase):
    def test_set_dict(self):
        d = {}
        protobuf.set(d, 'foo', 'bar')
        assert d == {'foo': 'bar'}

    def test_set_pb2(self):
        op = ops.Operation()
        protobuf.set(op, 'name', 'foo')
        assert op.name == 'foo'

    def test_set_nested(self):
        d = {}
        protobuf.set(d, 'foo.bar', 'baz')
        assert d == {'foo': {'bar': 'baz'}}

    def test_invalid_object(self):
        obj = object()
        with pytest.raises(TypeError):
            protobuf.set(obj, 'foo', 'bar')

    def test_set_list(self):
        list_ops_response = ops.ListOperationsResponse()
        protobuf.set(list_ops_response, 'operations', [
            {'name': 'foo'},
            ops.Operation(name='bar'),
        ])
        assert len(list_ops_response.operations) == 2
        for op in list_ops_response.operations:
            assert isinstance(op, ops.Operation)
        assert list_ops_response.operations[0].name == 'foo'
        assert list_ops_response.operations[1].name == 'bar'

    def test_set_list_clear_existing(self):
        list_ops_response = ops.ListOperationsResponse(
            operations=[{'name': 'baz'}],
        )
        protobuf.set(list_ops_response, 'operations', [
            {'name': 'foo'},
            ops.Operation(name='bar'),
        ])
        assert len(list_ops_response.operations) == 2
        for op in list_ops_response.operations:
            assert isinstance(op, ops.Operation)
        assert list_ops_response.operations[0].name == 'foo'
        assert list_ops_response.operations[1].name == 'bar'

    def test_set_dict_nested_with_message(self):
        rule = http_pb2.HttpRule()
        pattern = http_pb2.CustomHttpPattern(kind='foo', path='bar')
        protobuf.set(rule, 'custom', pattern)
        assert rule.custom.kind == 'foo'
        assert rule.custom.path == 'bar'

    def test_set_dict_nested_with_dict(self):
        rule = http_pb2.HttpRule()
        pattern = {'kind': 'foo', 'path': 'bar'}
        protobuf.set(rule, 'custom', pattern)
        assert rule.custom.kind == 'foo'
        assert rule.custom.path == 'bar'


class SetDefaultTests(unittest.TestCase):
    def test_dict_unset(self):
        d = {}
        protobuf.setdefault(d, 'foo', 'bar')
        assert d == {'foo': 'bar'}

    def test_dict_falsy(self):
        d = {'foo': None}
        protobuf.setdefault(d, 'foo', 'bar')
        assert d == {'foo': 'bar'}

    def test_dict_truthy(self):
        d = {'foo': 'bar'}
        protobuf.setdefault(d, 'foo', 'baz')
        assert d == {'foo': 'bar'}

    def test_pb2_falsy(self):
        op = ops.Operation()
        protobuf.setdefault(op, 'name', 'foo')
        assert op.name == 'foo'

    def test_pb2_truthy(self):
        op = ops.Operation(name='bar')
        protobuf.setdefault(op, 'name', 'foo')
        assert op.name == 'bar'



# Copyright 2016, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pylint: disable=missing-docstring,no-self-use,no-init,invalid-name
"""Unit tests for bundling."""

from __future__ import absolute_import

import mock
import unittest2

from google.gax import BundleDescriptor, BundleOptions, bundling

from tests.fixtures.fixture_pb2 import Bundled, Outer, Simple


def _Simple(value, other_value=None):
    if other_value is None:
        return Simple(field1=value)
    else:
        return Simple(field1=value, field2=other_value)


def _Outer(value):
    return Outer(inner=_Simple(value, value), field1=value)


def _Bundled(value):
    return Bundled(field1=value)


class TestComputeBundleId(unittest2.TestCase):

    def test_computes_bundle_ids_ok(self):
        tests = [
            {
                'message': 'single field value',
                'object': _Simple('dummy_value'),
                'fields': ['field1'],
                'want': ('dummy_value',)
            }, {
                'message': 'composite value with None',
                'object': _Simple('dummy_value'),
                'fields': ['field1', 'field2'],
                'want': ('dummy_value', None)
            }, {
                'message': 'a composite value',
                'object': _Simple('dummy_value', 'other_value'),
                'fields': ['field1', 'field2'],
                'want': ('dummy_value', 'other_value')
            }, {
                'message': 'a simple dotted value',
                'object': _Outer('this is dotty'),
                'fields': ['inner.field1'],
                'want': ('this is dotty',)
            }, {
                'message': 'a complex case',
                'object': _Outer('what!?'),
                'fields': ['inner.field1', 'inner.field2', 'field1'],
                'want': ('what!?', 'what!?', 'what!?')
            }
        ]
        for t in tests:
            got = bundling.compute_bundle_id(t['object'], t['fields'])
            message = 'failed while making an id for {}'.format(t['message'])
            self.assertEqual(got, t['want'], message)

    def test_should_raise_if_fields_are_missing(self):
        tests = [
            {
                'object': _Simple('dummy_value'),
                'fields': ['field3'],
            }, {
                'object': _Simple('dummy_value'),
                'fields': ['field1', 'field3'],
            }, {
                'object': _Simple('dummy_value', 'other_value'),
                'fields': ['field1', 'field3'],
            }, {
                'object': _Outer('this is dotty'),
                'fields': ['inner.field3'],
            }, {
                'object': _Outer('what!?'),
                'fields': ['inner.field4'],
            }
        ]
        for t in tests:
            self.assertRaises(AttributeError,
                              bundling.compute_bundle_id,
                              t['object'],
                              t['fields'])


def _return_request(req):
    """A dummy api call that simply returns the request."""
    return req


def _return_kwargs(dummy_req, **kwargs):
    """A dummy api call that simply returns its keyword arguments."""
    return kwargs


def _make_a_test_task(api_call=_return_request):
    return bundling.Task(
        api_call,
        'an_id',
        'field1',
        _Bundled([]),
        dict())


def _extend_with_n_elts(a_task, elt, n):
    return a_task.extend([elt] * n)


def _raise_exc(dummy_req):
    """A dummy api call that raises an exception"""
    raise ValueError('Raised in a test')


class TestTask(unittest2.TestCase):

    def test_extend_increases_the_element_count(self):
        simple_msg = 'a simple msg'
        tests = [
            {
                'update': (lambda t: None),
                'message': 'no messages added',
                'want': 0
            }, {
                'update': (lambda t: t.extend([simple_msg])),
                'message': 'a single message added',
                'want': 1
            }, {
                'update': (lambda t: _extend_with_n_elts(t, simple_msg, 5)),
                'message': 'a 5 messages added',
                'want': 5
            }
        ]
        for t in tests:
            test_task = _make_a_test_task()
            t['update'](test_task)
            got = test_task.element_count
            message = 'bad message count when {}'.format(t['message'])
            self.assertEqual(got, t['want'], message)

    def test_extend_increases_the_request_byte_count(self):
        simple_msg = 'a simple msg'
        tests = [
            {
                'update': (lambda t: None),
                'message': 'no messages added',
                'want': 0
            }, {
                'update': (lambda t: t.extend([simple_msg])),
                'message': 'a single bundle message',
                'want': len(simple_msg)
            }, {
                'update': (lambda t: _extend_with_n_elts(t, simple_msg, 5)),
                'message': '5 bundled messages',
                'want': 5 * len(simple_msg)
            }
        ]
        for t in tests:
            test_task = _make_a_test_task()
            t['update'](test_task)
            got = test_task.request_bytesize
            message = 'bad message count when {}'.format(t['message'])
            self.assertEqual(got, t['want'], message)

    def test_run_sends_the_bundle_elements(self):
        simple_msg = 'a simple msg'
        tests = [
            {
                'update': (lambda t: None),
                'message': 'no messages added',
                'has_event': False,
                'count_before_run': 0,
                'want': []
            }, {
                'update': (lambda t: t.extend([simple_msg])),
                'message': 'a single bundled message',
                'has_event': True,
                'count_before_run': 1,
                'want': _Bundled([simple_msg])
            }, {
                'update': (lambda t: _extend_with_n_elts(t, simple_msg, 5)),
                'message': '5 bundle messages',
                'has_event': True,
                'count_before_run': 5,
                'want': _Bundled([simple_msg] * 5)
            }
        ]
        for t in tests:
            test_task = _make_a_test_task()
            event = t['update'](test_task)
            self.assertEqual(test_task.element_count, t['count_before_run'])
            test_task.run()
            self.assertEqual(test_task.element_count, 0)
            self.assertEqual(test_task.request_bytesize, 0)
            if t['has_event']:
                self.assertIsNotNone(
                    event,
                    'expected event for {}'.format(t['message']))
                got = event.result
                message = 'bad output when run with {}'.format(t['message'])
                self.assertEqual(got, t['want'], message)

    def test_run_adds_an_error_if_execution_fails(self):
        simple_msg = 'a simple msg'
        test_task = _make_a_test_task(api_call=_raise_exc)
        event = test_task.extend([simple_msg])
        self.assertEqual(test_task.element_count, 1)
        test_task.run()
        self.assertEqual(test_task.element_count, 0)
        self.assertEqual(test_task.request_bytesize, 0)
        self.assertTrue(isinstance(event.result, ValueError))

    def test_calling_the_canceller_stops_the_element_from_getting_sent(self):
        an_elt = 'a simple msg'
        another_msg = 'another msg'
        test_task = _make_a_test_task()
        an_event = test_task.extend([an_elt])
        another_event = test_task.extend([another_msg])
        self.assertEqual(test_task.element_count, 2)
        self.assertTrue(an_event.cancel())
        self.assertEqual(test_task.element_count, 1)
        self.assertFalse(an_event.cancel())
        self.assertEqual(test_task.element_count, 1)
        test_task.run()
        self.assertEqual(test_task.element_count, 0)
        self.assertEqual(_Bundled([another_msg]), another_event.result)
        self.assertFalse(an_event.is_set())
        self.assertIsNone(an_event.result)


SIMPLE_DESCRIPTOR = BundleDescriptor('field1', [])
DEMUX_DESCRIPTOR = BundleDescriptor('field1', [], subresponse_field='field1')


class TestExecutor(unittest2.TestCase):

    def test_api_calls_are_grouped_by_bundle_id(self):
        an_elt = 'dummy message'
        api_call = _return_request
        bundle_ids = ['id1', 'id2']
        threshold = 5  # arbitrary
        options = BundleOptions(element_count_threshold=threshold)
        bundler = bundling.Executor(options)
        for an_id in bundle_ids:
            for i in range(threshold - 1):
                got_event = bundler.schedule(
                    api_call,
                    an_id,
                    SIMPLE_DESCRIPTOR,
                    _Bundled([an_elt])
                )
                self.assertIsNotNone(
                    got_event.canceller,
                    'missing canceller after element #{}'.format(i))
                self.assertFalse(
                    got_event.is_set(),
                    'event unexpectedly set after element #{}'.format(i))
                self.assertIsNone(got_event.result)
        for an_id in bundle_ids:
            got_event = bundler.schedule(
                api_call,
                an_id,
                SIMPLE_DESCRIPTOR,
                _Bundled([an_elt])
            )
            self.assertIsNotNone(got_event.canceller,
                                 'missing expected canceller')
            self.assertTrue(
                got_event.is_set(),
                'event is not set after triggering element')
            self.assertEqual(_Bundled([an_elt] * threshold),
                             got_event.result)

    def test_each_event_has_exception_when_demuxed_api_call_fails(self):
        an_elt = 'dummy message'
        api_call = _raise_exc
        bundle_id = 'an_id'
        threshold = 5  # arbitrary, greater than 1
        options = BundleOptions(element_count_threshold=threshold)
        bundler = bundling.Executor(options)
        events = []
        for i in range(threshold - 1):
            got_event = bundler.schedule(
                api_call,
                bundle_id,
                DEMUX_DESCRIPTOR,
                _Bundled(['%s%d' % (an_elt, i)])
            )
            self.assertFalse(
                got_event.is_set(),
                'event unexpectedly set after element #{}'.format(i))
            self.assertIsNone(got_event.result)
            events.append(got_event)
        last_event = bundler.schedule(
            api_call,
            bundle_id,
            DEMUX_DESCRIPTOR,
            _Bundled(['%s%d' % (an_elt, threshold - 1)])
        )
        events.append(last_event)

        previous_event = None
        for event in events:
            if previous_event:
                self.assertTrue(previous_event != event)
            self.assertTrue(event.is_set(),
                            'event is not set after triggering element')
            self.assertTrue(isinstance(event.result, ValueError))
            previous_event = event

    def test_each_event_has_its_result_from_a_demuxed_api_call(self):
        an_elt = 'dummy message'
        api_call = _return_request
        bundle_id = 'an_id'
        threshold = 5  # arbitrary, greater than 1
        options = BundleOptions(element_count_threshold=threshold)
        bundler = bundling.Executor(options)
        events = []

        # send 3 groups of elements of different sizes in the bundle
        for i in range(1, 4):
            got_event = bundler.schedule(
                api_call,
                bundle_id,
                DEMUX_DESCRIPTOR,
                _Bundled(['%s%d' % (an_elt, i)] * i)
            )
            events.append(got_event)
        previous_event = None
        for i, event in enumerate(events):
            index = i + 1
            if previous_event:
                self.assertTrue(previous_event != event)
            self.assertTrue(event.is_set(),
                            'event is not set after triggering element')
            self.assertEqual(event.result,
                             _Bundled(['%s%d' % (an_elt, index)] * index))
            previous_event = event

    def test_each_event_has_same_result_from_mismatched_demuxed_api_call(self):
        an_elt = 'dummy message'
        mismatched_result = _Bundled([an_elt, an_elt])
        bundle_id = 'an_id'
        threshold = 5  # arbitrary, greater than 1
        options = BundleOptions(element_count_threshold=threshold)
        bundler = bundling.Executor(options)
        events = []

        # send 3 groups of elements of different sizes in the bundle
        for i in range(1, 4):
            got_event = bundler.schedule(
                lambda x: mismatched_result,
                bundle_id,
                DEMUX_DESCRIPTOR,
                _Bundled(['%s%d' % (an_elt, i)] * i)
            )
            events.append(got_event)
        previous_event = None
        for i, event in enumerate(events):
            if previous_event:
                self.assertTrue(previous_event != event)
            self.assertTrue(event.is_set(),
                            'event is not set after triggering element')
            self.assertEqual(event.result, mismatched_result)
            previous_event = event

    def test_schedule_passes_kwargs(self):
        an_elt = 'dummy_msg'
        options = BundleOptions(element_count_threshold=1)
        bundle_id = 'an_id'
        bundler = bundling.Executor(options)
        event = bundler.schedule(
            _return_kwargs,
            bundle_id,
            SIMPLE_DESCRIPTOR,
            _Bundled([an_elt]),
            {'an_option': 'a_value'}
        )
        self.assertEqual('a_value',
                         event.result['an_option'])


class TestExecutor_ElementCountTrigger(unittest2.TestCase):

    def test_api_call_not_invoked_until_threshold(self):
        an_elt = 'dummy message'
        an_id = 'bundle_id'
        api_call = _return_request
        threshold = 3  # arbitrary
        options = BundleOptions(element_count_threshold=threshold)
        bundler = bundling.Executor(options)
        for i in range(threshold):
            got_event = bundler.schedule(
                api_call,
                an_id,
                SIMPLE_DESCRIPTOR,
                _Bundled([an_elt])
            )
            self.assertIsNotNone(
                got_event.canceller,
                'missing canceller after element #{}'.format(i))
            if i + 1 < threshold:
                self.assertFalse(got_event.is_set())
                self.assertIsNone(got_event.result)
            else:
                self.assertTrue(got_event.is_set())
                self.assertEqual(_Bundled([an_elt] * threshold),
                                 got_event.result)


class TestExecutor_RequestByteTrigger(unittest2.TestCase):

    def test_api_call_not_invoked_until_threshold(self):
        an_elt = 'dummy message'
        an_id = 'bundle_id'
        api_call = _return_request
        elts_for_threshold = 3
        threshold = elts_for_threshold * len(an_elt)  # arbitrary
        options = BundleOptions(request_byte_threshold=threshold)
        bundler = bundling.Executor(options)
        for i in range(elts_for_threshold):
            got_event = bundler.schedule(
                api_call,
                an_id,
                SIMPLE_DESCRIPTOR,
                _Bundled([an_elt])
            )
            self.assertIsNotNone(
                got_event.canceller,
                'missing canceller after element #{}'.format(i))
            if i + 1 < elts_for_threshold:
                self.assertFalse(got_event.is_set())
                self.assertIsNone(got_event.result)
            else:
                self.assertTrue(got_event.is_set())
                self.assertEqual(_Bundled([an_elt] * elts_for_threshold),
                                 got_event.result)


class TestExecutor_DelayThreshold(unittest2.TestCase):

    @mock.patch('google.gax.bundling.TIMER_FACTORY')
    def test_api_call_is_scheduled_on_timer(self, timer_class):
        an_elt = 'dummy message'
        an_id = 'bundle_id'
        api_call = _return_request
        delay_threshold = 3
        options = BundleOptions(delay_threshold=delay_threshold)
        bundler = bundling.Executor(options)
        got_event = bundler.schedule(
            api_call,
            an_id,
            SIMPLE_DESCRIPTOR,
            _Bundled([an_elt])
        )
        self.assertIsNotNone(got_event, 'missing event after first request')
        self.assertIsNone(got_event.result)
        self.assertTrue(timer_class.called)
        timer_args, timer_kwargs = timer_class.call_args_list[0]
        self.assertEqual(delay_threshold, timer_args[0])
        self.assertEqual({'args': [an_id]}, timer_kwargs)
        timer_class.return_value.start.assert_called_once_with()


class TestEvent(unittest2.TestCase):

    def test_can_be_set(self):
        ev = bundling.Event()
        self.assertFalse(ev.is_set())
        ev.set()
        self.assertTrue(ev.is_set())

    def test_can_be_cleared(self):
        ev = bundling.Event()
        ev.result = object()
        ev.set()
        self.assertTrue(ev.is_set())
        self.assertIsNotNone(ev.result)
        ev.clear()
        self.assertFalse(ev.is_set())
        self.assertIsNone(ev.result)

    def test_cancel_returns_false_without_canceller(self):
        ev = bundling.Event()
        self.assertFalse(ev.cancel())

    def test_cancel_returns_canceller_result(self):
        ev = bundling.Event()
        ev.canceller = lambda: True
        self.assertTrue(ev.cancel())
        ev.canceller = lambda: False
        self.assertFalse(ev.cancel())

    def test_wait_does_not_block_if_event_is_set(self):
        ev = bundling.Event()
        ev.set()
        self.assertTrue(ev.wait())



# Copyright 2016, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Provides GAX exceptions."""

from __future__ import absolute_import

from google.gax import config


class GaxError(Exception):
    """Common base class for exceptions raised by GAX.

    Attributes:
      msg (string): describes the error that occurred.
      cause (Exception, optional): the exception raised by a lower
        layer of the RPC stack (for example, gRPC) that caused this
        exception, or None if this exception originated in GAX.
    """
    def __init__(self, msg, cause=None):
        super(GaxError, self).__init__(msg)
        self.cause = cause

    def __str__(self):
        msg = super(GaxError, self).__str__()
        if not self.cause:
            return msg

        return 'GaxError({}, caused by {})'.format(msg, self.cause)


def create_error(msg, cause=None):
    """Creates an error.

    Uses a Python built-in exception if one is available, and a
    GaxError otherwise.

    Attributes:
      msg (string): describes the error that occurred.
      cause (Exception, optional): the exception raised by a lower
        layer of the RPC stack (for example, gRPC) that caused this
        exception, or None if this exception originated in GAX.
    """
    if config.NAME_STATUS_CODES.get(
            config.exc_to_code(cause)) == 'INVALID_ARGUMENT':
        return ValueError('{}: {}'.format(msg, cause))
    else:
        return GaxError(msg, cause)


class RetryError(GaxError):
    """Indicates an error during automatic GAX retrying."""
    pass



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pylint: disable=missing-docstring,no-self-use,no-init,invalid-name
"""Unit tests for the path_template module."""

from __future__ import absolute_import

import unittest2

from google.gax.path_template import PathTemplate, ValidationException


class TestPathTemplate(unittest2.TestCase):
    """Unit tests for PathTemplate."""

    def test_len(self):
        self.assertEqual(len(PathTemplate('a/b/**/*/{a=hello/world}')), 6)

    def test_fail_invalid_token(self):
        self.assertRaises(ValidationException,
                          PathTemplate, 'hello/wor*ld')

    def test_fail_when_impossible_match(self):
        template = PathTemplate('hello/world')
        self.assertRaises(ValidationException,
                          template.match, 'hello')
        template = PathTemplate('hello/world')
        self.assertRaises(ValidationException,
                          template.match, 'hello/world/fail')

    def test_fail_mismatched_literal(self):
        template = PathTemplate('hello/world')
        self.assertRaises(ValidationException,
                          template.match, 'hello/world2')

    def test_fail_when_multiple_path_wildcards(self):
        self.assertRaises(ValidationException,
                          PathTemplate, 'buckets/*/**/**/objects/*')

    def test_fail_if_inner_binding(self):
        self.assertRaises(ValidationException,
                          PathTemplate, 'buckets/{hello={world}}')

    def test_fail_unexpected_eof(self):
        self.assertRaises(ValidationException,
                          PathTemplate, 'a/{hello=world')

    def test_match_atomic_resource_name(self):
        template = PathTemplate('buckets/*/*/objects/*')
        self.assertEqual({'$0': 'f', '$1': 'o', '$2': 'bar'},
                         template.match('buckets/f/o/objects/bar'))
        template = PathTemplate('/buckets/{hello}')
        self.assertEqual({'hello': 'world'},
                         template.match('buckets/world'))
        template = PathTemplate('/buckets/{hello=*}')
        self.assertEqual({'hello': 'world'},
                         template.match('buckets/world'))

    def test_match_escaped_chars(self):
        template = PathTemplate('buckets/*/objects')
        self.assertEqual({'$0': 'hello%2F%2Bworld'},
                         template.match('buckets/hello%2F%2Bworld/objects'))

    def test_match_template_with_unbounded_wildcard(self):
        template = PathTemplate('buckets/*/objects/**')
        self.assertEqual({'$0': 'foo', '$1': 'bar/baz'},
                         template.match('buckets/foo/objects/bar/baz'))

    def test_match_with_unbound_in_middle(self):
        template = PathTemplate('bar/**/foo/*')
        self.assertEqual({'$0': 'foo/foo', '$1': 'bar'},
                         template.match('bar/foo/foo/foo/bar'))

    def test_render_atomic_resource(self):
        template = PathTemplate('buckets/*/*/*/objects/*')
        url = template.render({
            '$0': 'f', '$1': 'o', '$2': 'o', '$3': 'google.com:a-b'})
        self.assertEqual(url, 'buckets/f/o/o/objects/google.com:a-b')

    def test_render_fail_when_too_few_variables(self):
        template = PathTemplate('buckets/*/*/*/objects/*')
        self.assertRaises(ValidationException,
                          template.render,
                          {'$0': 'f', '$1': 'l', '$2': 'o'})

    def test_render_with_unbound_in_middle(self):
        template = PathTemplate('bar/**/foo/*')
        url = template.render({'$0': '1/2', '$1': '3'})
        self.assertEqual(url, 'bar/1/2/foo/3')

    def test_to_string(self):
        template = PathTemplate('bar/**/foo/*')
        self.assertEqual(str(template), 'bar/{$0=**}/foo/{$1=*}')
        template = PathTemplate('buckets/*/objects/*')
        self.assertEqual(str(template), 'buckets/{$0=*}/objects/{$1=*}')
        template = PathTemplate('/buckets/{hello}')
        self.assertEqual(str(template), 'buckets/{hello=*}')
        template = PathTemplate('/buckets/{hello=what}/{world}')
        self.assertEqual(str(template), 'buckets/{hello=what}/{world=*}')
        template = PathTemplate('/buckets/helloazAZ09-.~_what')
        self.assertEqual(str(template), 'buckets/helloazAZ09-.~_what')



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pylint: disable=missing-docstring,no-self-use,no-init,invalid-name,protected-access
"""Unit tests for api_callable"""

from __future__ import absolute_import, division

import platform

import grpc
import mock
import pkg_resources
import unittest2

from google.gax import (
    __version__ as GAX_VERSION, _CallSettings, api_callable, BackoffSettings,
    BundleDescriptor, BundleOptions, bundling, CallOptions, INITIAL_PAGE,
    PageDescriptor, RetryOptions)
from google.gax.errors import GaxError

# pylint: disable=no-member
GRPC_VERSION = pkg_resources.get_distribution('grpcio').version
# pylint: enable=no-member

_SERVICE_NAME = 'test.interface.v1.api'


_A_CONFIG = {
    'interfaces': {
        _SERVICE_NAME: {
            'retry_codes': {
                'foo_retry': ['code_a', 'code_b'],
                'bar_retry': ['code_c']
            },
            'retry_params': {
                'default': {
                    'initial_retry_delay_millis': 100,
                    'retry_delay_multiplier': 1.2,
                    'max_retry_delay_millis': 1000,
                    'initial_rpc_timeout_millis': 300,
                    'rpc_timeout_multiplier': 1.3,
                    'max_rpc_timeout_millis': 3000,
                    'total_timeout_millis': 30000
                }
            },
            'methods': {
                # Note that GAX should normalize this to snake case
                'BundlingMethod': {
                    'retry_codes_name': 'foo_retry',
                    'retry_params_name': 'default',
                    'timeout_millis': 25000,
                    'bundling': {
                        'element_count_threshold': 6,
                        'element_count_limit': 10
                    }
                },
                'PageStreamingMethod': {
                    'retry_codes_name': 'bar_retry',
                    'retry_params_name': 'default',
                    'timeout_millis': 12000
                }
            }
        }
    }
}


_PAGE_DESCRIPTORS = {
    'page_streaming_method': PageDescriptor(
        'page_token', 'next_page_token', 'page_streams')
}


_BUNDLE_DESCRIPTORS = {'bundling_method': BundleDescriptor('bundled_field', [])}


_RETRY_DICT = {'code_a': Exception,
               'code_b': Exception,
               'code_c': Exception}


_FAKE_STATUS_CODE_1 = object()


class CustomException(Exception):
    def __init__(self, msg, code):
        super(CustomException, self).__init__(msg)
        self.code = code


class AnotherException(Exception):
    pass


class TestCreateApiCallable(unittest2.TestCase):

    def test_call_api_call(self):
        settings = _CallSettings()
        my_callable = api_callable.create_api_call(
            lambda _req, _timeout: 42, settings)
        self.assertEqual(my_callable(None), 42)

    def test_call_override(self):
        settings = _CallSettings(timeout=10)
        my_callable = api_callable.create_api_call(
            lambda _req, timeout: timeout, settings)
        self.assertEqual(my_callable(None, CallOptions(timeout=20)), 20)

    def test_call_kwargs(self):
        settings = _CallSettings(kwargs={'key': 'value'})
        my_callable = api_callable.create_api_call(
            lambda _req, _timeout, **kwargs: kwargs['key'], settings)
        self.assertEqual(my_callable(None), 'value')
        self.assertEqual(my_callable(None, CallOptions(key='updated')),
                         'updated')

    @mock.patch('time.time')
    @mock.patch('google.gax.config.exc_to_code')
    def test_retry(self, mock_exc_to_code, mock_time):
        mock_exc_to_code.side_effect = lambda e: e.code
        to_attempt = 3
        retry = RetryOptions(
            [_FAKE_STATUS_CODE_1],
            BackoffSettings(0, 0, 0, 0, 0, 0, 1))

        # Succeeds on the to_attempt'th call, and never again afterward
        mock_call = mock.Mock()
        mock_call.side_effect = ([CustomException('', _FAKE_STATUS_CODE_1)] *
                                 (to_attempt - 1) + [mock.DEFAULT])
        mock_call.return_value = 1729
        mock_time.return_value = 0
        settings = _CallSettings(timeout=0, retry=retry)
        my_callable = api_callable.create_api_call(mock_call, settings)
        self.assertEqual(my_callable(None), 1729)
        self.assertEqual(mock_call.call_count, to_attempt)

    def test_page_streaming(self):
        # A mock grpc function that page streams a list of consecutive
        # integers, returning `page_size` integers with each call and using
        # the next integer to return as the page token, until `pages_to_stream`
        # pages have been returned.
        # pylint:disable=too-many-locals
        page_size = 3
        pages_to_stream = 5

        # pylint: disable=abstract-method, too-few-public-methods
        class PageStreamingRequest(object):
            def __init__(self, page_token=0):
                self.page_token = page_token

        class PageStreamingResponse(object):
            def __init__(self, nums=(), next_page_token=0):
                self.nums = nums
                self.next_page_token = next_page_token

        fake_grpc_func_descriptor = PageDescriptor(
            'page_token', 'next_page_token', 'nums')

        def grpc_return_value(request, *dummy_args, **dummy_kwargs):
            start = int(request.page_token)
            if start > 0 and start < page_size * pages_to_stream:
                return PageStreamingResponse(
                    nums=list(range(start,
                                    start + page_size)),
                    next_page_token=start + page_size)
            elif start >= page_size * pages_to_stream:
                return PageStreamingResponse()
            else:
                return PageStreamingResponse(nums=list(range(page_size)),
                                             next_page_token=page_size)

        with mock.patch('grpc.UnaryUnaryMultiCallable') as mock_grpc:
            mock_grpc.side_effect = grpc_return_value
            settings = _CallSettings(
                page_descriptor=fake_grpc_func_descriptor, timeout=0)
            my_callable = api_callable.create_api_call(
                mock_grpc, settings=settings)
            self.assertEqual(list(my_callable(PageStreamingRequest())),
                             list(range(page_size * pages_to_stream)))

            unflattened_option = CallOptions(page_token=INITIAL_PAGE)
            # Expect a list of pages_to_stream pages, each of size page_size,
            # plus one empty page
            expected = [list(range(page_size * n, page_size * (n + 1)))
                        for n in range(pages_to_stream)] + [()]
            self.assertEqual(list(my_callable(PageStreamingRequest(),
                                              unflattened_option)),
                             expected)

            pages_already_read = 2
            explicit_page_token_option = CallOptions(
                page_token=str(page_size * pages_already_read))
            # Expect a list of pages_to_stream pages, each of size page_size,
            # plus one empty page, minus the pages_already_read
            expected = [list(range(page_size * n, page_size * (n + 1)))
                        for n in range(pages_already_read, pages_to_stream)]
            expected += [()]
            self.assertEqual(list(my_callable(PageStreamingRequest(),
                                              explicit_page_token_option)),
                             expected)

    def test_bundling_page_streaming_error(self):
        settings = _CallSettings(
            page_descriptor=object(), bundle_descriptor=object(),
            bundler=object())
        with self.assertRaises(ValueError):
            api_callable.create_api_call(lambda _req, _timeout: 42, settings)

    def test_bundling(self):
        # pylint: disable=abstract-method, too-few-public-methods
        class BundlingRequest(object):
            def __init__(self, elements=None):
                self.elements = elements

        fake_grpc_func_descriptor = BundleDescriptor('elements', [])
        bundler = bundling.Executor(BundleOptions(element_count_threshold=8))

        def my_func(request, dummy_timeout):
            return len(request.elements)

        settings = _CallSettings(
            bundler=bundler, bundle_descriptor=fake_grpc_func_descriptor,
            timeout=0)
        my_callable = api_callable.create_api_call(my_func, settings)
        first = my_callable(BundlingRequest([0] * 3))
        self.assertIsInstance(first, bundling.Event)
        self.assertIsNone(first.result)  # pylint: disable=no-member
        second = my_callable(BundlingRequest([0] * 5))
        self.assertEqual(second.result, 8)  # pylint: disable=no-member

    def test_construct_settings(self):
        defaults = api_callable.construct_settings(
            _SERVICE_NAME, _A_CONFIG, dict(), _RETRY_DICT,
            bundle_descriptors=_BUNDLE_DESCRIPTORS,
            page_descriptors=_PAGE_DESCRIPTORS,
            kwargs={'key1': 'value1'})

        # Check values of the bundling method settings.
        settings = defaults['bundling_method']
        self.assertAlmostEqual(settings.timeout, 25.0)
        self.assertIsInstance(settings.bundler, bundling.Executor)
        self.assertIsInstance(settings.bundle_descriptor, BundleDescriptor)
        self.assertIsNone(settings.page_descriptor)

        # Check values of the page streaming method settings.
        settings = defaults['page_streaming_method']
        self.assertAlmostEqual(settings.timeout, 12.0)
        self.assertIsNone(settings.bundler)
        self.assertIsNone(settings.bundle_descriptor)
        self.assertIsInstance(settings.page_descriptor, PageDescriptor)

        # Check values that should be the same in both method settings.
        for settings in defaults.values():
            self.assertIsInstance(settings.retry, RetryOptions)
            self.assertEqual(settings.kwargs['key1'], 'value1')
            self.assertIn('metadata', settings.kwargs)
            metadata = settings.kwargs['metadata'][0][1]
            self.assertIn('gl-python/%s' % platform.python_version(), metadata)
            self.assertIn('gax/%s' % GAX_VERSION, metadata)
            self.assertIn('grpc/%s' % GRPC_VERSION, metadata)

    def test_construct_with_explicit_metadata(self):
        defaults = api_callable.construct_settings(
            _SERVICE_NAME, _A_CONFIG, dict(), _RETRY_DICT,
            bundle_descriptors=_BUNDLE_DESCRIPTORS,
            page_descriptors=_PAGE_DESCRIPTORS,
            kwargs={'metadata': [('x-goog-api-client', 'foo/1.2.3')]})
        settings = defaults['bundling_method']
        metadata = settings.kwargs['metadata'][0][1]

        # Ensure that the metadata string contains items that GAX adds,
        # but not the original header.
        self.assertNotIn('foo/1.2.3', metadata)
        self.assertIn('gl-python/%s' % platform.python_version(), metadata)
        self.assertIn('gax/%s' % GAX_VERSION, metadata)
        self.assertIn('grpc/%s' % GRPC_VERSION, metadata)

    def test_construct_settings_override(self):
        _override = {
            'interfaces': {
                _SERVICE_NAME: {
                    'methods': {
                        'PageStreamingMethod': None,
                        'BundlingMethod': {
                            'timeout_millis': 8000,
                            'bundling': None
                        }
                    }
                }
            }
        }
        defaults = api_callable.construct_settings(
            _SERVICE_NAME, _A_CONFIG, _override, _RETRY_DICT,
            bundle_descriptors=_BUNDLE_DESCRIPTORS,
            page_descriptors=_PAGE_DESCRIPTORS)
        settings = defaults['bundling_method']
        self.assertAlmostEqual(settings.timeout, 8.0)
        self.assertIsNone(settings.bundler)
        self.assertIsNone(settings.page_descriptor)
        settings = defaults['page_streaming_method']
        self.assertAlmostEqual(settings.timeout, 12.0)
        self.assertIsInstance(settings.page_descriptor, PageDescriptor)
        self.assertIsNone(settings.retry)

    def test_construct_settings_override2(self):
        _override = {
            'interfaces': {
                _SERVICE_NAME: {
                    'retry_codes': {
                        'bar_retry': [],
                        'baz_retry': ['code_a']
                    },
                    'retry_params': {
                        'default': {
                            'initial_retry_delay_millis': 1000,
                            'retry_delay_multiplier': 1.2,
                            'max_retry_delay_millis': 10000,
                            'initial_rpc_timeout_millis': 3000,
                            'rpc_timeout_multiplier': 1.3,
                            'max_rpc_timeout_millis': 30000,
                            'total_timeout_millis': 300000
                        },
                    },
                    'methods': {
                        'BundlingMethod': {
                            'retry_params_name': 'default',
                            'retry_codes_name': 'baz_retry',
                        },
                    },
                }
            }
        }
        defaults = api_callable.construct_settings(
            _SERVICE_NAME, _A_CONFIG, _override, _RETRY_DICT,
            bundle_descriptors=_BUNDLE_DESCRIPTORS,
            page_descriptors=_PAGE_DESCRIPTORS)
        settings = defaults['bundling_method']
        backoff = settings.retry.backoff_settings
        self.assertEqual(backoff.initial_retry_delay_millis, 1000)
        self.assertEqual(settings.retry.retry_codes, [_RETRY_DICT['code_a']])
        self.assertIsInstance(settings.bundler, bundling.Executor)
        self.assertIsInstance(settings.bundle_descriptor, BundleDescriptor)

        # page_streaming_method is unaffected because it's not specified in
        # overrides. 'bar_retry' or 'default' definitions in overrides should
        # not affect the methods which are not in the overrides.
        settings = defaults['page_streaming_method']
        backoff = settings.retry.backoff_settings
        self.assertEqual(backoff.initial_retry_delay_millis, 100)
        self.assertEqual(backoff.retry_delay_multiplier, 1.2)
        self.assertEqual(backoff.max_retry_delay_millis, 1000)
        self.assertEqual(settings.retry.retry_codes, [_RETRY_DICT['code_c']])

    @mock.patch('google.gax.config.API_ERRORS', (CustomException, ))
    def test_catch_error(self):
        def abortion_error_func(*dummy_args, **dummy_kwargs):
            raise CustomException(None, None)

        def other_error_func(*dummy_args, **dummy_kwargs):
            raise AnotherException

        gax_error_callable = api_callable.create_api_call(
            abortion_error_func, _CallSettings())
        self.assertRaises(GaxError, gax_error_callable, None)

        other_error_callable = api_callable.create_api_call(
            other_error_func, _CallSettings())
        self.assertRaises(AnotherException, other_error_callable, None)

    def test_wrap_value_error(self):

        invalid_attribute_exc = grpc.RpcError()
        invalid_attribute_exc.code = lambda: grpc.StatusCode.INVALID_ARGUMENT

        def value_error_func(*dummy_args, **dummy_kwargs):
            raise invalid_attribute_exc

        value_error_callable = api_callable.create_api_call(
            value_error_func, _CallSettings())
        self.assertRaises(ValueError, value_error_callable, None)



# Copyright 2016, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Provides gRPC authentication support using google.auth."""

from __future__ import absolute_import

import google.auth
import google.auth.credentials
import google.auth.transport.grpc

try:
    import google.auth.transport.requests
    # pylint: disable=invalid-name
    # Pylint recognizes this as a class, but we're aliasing it as a factory
    # function.
    _request_factory = google.auth.transport.requests.Request
    # pylint: enable=invalid-name
except ImportError:
    try:
        import httplib2
        import google_auth_httplib2

        def _request_factory():
            http = httplib2.Http()
            return google_auth_httplib2.Request(http)

    except ImportError:
        raise ImportError(
            'No HTTP transport is available. Please install requests or '
            'httplib2 and google_auth_httplib2.')


def get_default_credentials(scopes):
    """Gets the Application Default Credentials."""
    credentials, _ = google.auth.default(scopes=scopes)
    return credentials


def secure_authorized_channel(
        credentials, target, ssl_credentials=None):
    """Creates a secure authorized gRPC channel."""
    http_request = _request_factory()
    return google.auth.transport.grpc.secure_authorized_channel(
        credentials, http_request, target,
        ssl_credentials=ssl_credentials)



# Copyright 2017, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pylint: disable=missing-docstring,invalid-name
"""Unit tests for retry"""

from __future__ import absolute_import, division

import mock
import unittest2

from google.gax import BackoffSettings, errors, retry, RetryOptions

_MILLIS_PER_SEC = 1000


_FAKE_STATUS_CODE_1 = object()


_FAKE_STATUS_CODE_2 = object()


class CustomException(Exception):
    def __init__(self, msg, code):
        super(CustomException, self).__init__(msg)
        self.code = code


class TestRetry(unittest2.TestCase):

    @mock.patch('google.gax.config.exc_to_code')
    @mock.patch('time.time')
    def test_retryable_without_timeout(self, mock_time, mock_exc_to_code):
        mock_time.return_value = 0
        mock_exc_to_code.side_effect = lambda e: e.code

        to_attempt = 3
        mock_call = mock.Mock()
        mock_call.side_effect = ([CustomException('', _FAKE_STATUS_CODE_1)] *
                                 (to_attempt - 1) + [mock.DEFAULT])
        mock_call.return_value = 1729

        retry_options = RetryOptions(
            [_FAKE_STATUS_CODE_1],
            BackoffSettings(0, 0, 0, None, None, None, None))

        my_callable = retry.retryable(mock_call, retry_options)

        self.assertEqual(my_callable(None), 1729)
        self.assertEqual(to_attempt, mock_call.call_count)

    @mock.patch('google.gax.config.exc_to_code')
    @mock.patch('time.time')
    def test_retryable_with_timeout(self, mock_time, mock_exc_to_code):
        mock_time.return_value = 1
        mock_exc_to_code.side_effect = lambda e: e.code

        mock_call = mock.Mock()
        mock_call.side_effect = [CustomException('', _FAKE_STATUS_CODE_1),
                                 mock.DEFAULT]
        mock_call.return_value = 1729

        retry_options = RetryOptions(
            [_FAKE_STATUS_CODE_1],
            BackoffSettings(0, 0, 0, 0, 0, 0, 0))

        my_callable = retry.retryable(mock_call, retry_options)

        self.assertRaises(errors.RetryError, my_callable)
        self.assertEqual(0, mock_call.call_count)

    @mock.patch('google.gax.config.exc_to_code')
    @mock.patch('time.time')
    def test_retryable_when_no_codes(self, mock_time, mock_exc_to_code):
        mock_time.return_value = 0
        mock_exc_to_code.side_effect = lambda e: e.code

        mock_call = mock.Mock()
        mock_call.side_effect = [CustomException('', _FAKE_STATUS_CODE_1),
                                 mock.DEFAULT]
        mock_call.return_value = 1729

        retry_options = RetryOptions(
            [],
            BackoffSettings(0, 0, 0, 0, 0, 0, 1))

        my_callable = retry.retryable(mock_call, retry_options)

        try:
            my_callable(None)
            self.fail('Should not have been reached')
        except errors.RetryError as exc:
            self.assertIsInstance(exc.cause, CustomException)

        self.assertEqual(1, mock_call.call_count)

    @mock.patch('google.gax.config.exc_to_code')
    @mock.patch('time.time')
    def test_retryable_aborts_on_unexpected_exception(
            self, mock_time, mock_exc_to_code):
        mock_time.return_value = 0
        mock_exc_to_code.side_effect = lambda e: e.code

        mock_call = mock.Mock()
        mock_call.side_effect = [CustomException('', _FAKE_STATUS_CODE_2),
                                 mock.DEFAULT]
        mock_call.return_value = 1729

        retry_options = RetryOptions(
            [_FAKE_STATUS_CODE_1],
            BackoffSettings(0, 0, 0, 0, 0, 0, 1))

        my_callable = retry.retryable(mock_call, retry_options)

        try:
            my_callable(None)
            self.fail('Should not have been reached')
        except errors.RetryError as exc:
            self.assertIsInstance(exc.cause, CustomException)

        self.assertEqual(1, mock_call.call_count)

    @mock.patch('google.gax.config.exc_to_code')
    @mock.patch('time.sleep')
    @mock.patch('time.time')
    def test_retryable_exponential_backoff(
            self, mock_time, mock_sleep, mock_exc_to_code):
        def incr_time(secs):
            mock_time.return_value += secs

        def api_call(timeout):
            incr_time(timeout)
            raise CustomException(str(timeout), _FAKE_STATUS_CODE_1)

        mock_time.return_value = 0
        mock_sleep.side_effect = incr_time
        mock_exc_to_code.side_effect = lambda e: e.code

        mock_call = mock.Mock()
        mock_call.side_effect = api_call

        params = BackoffSettings(3, 2, 24, 5, 2, 80, 2500)
        retry_options = RetryOptions([_FAKE_STATUS_CODE_1], params)

        my_callable = retry.retryable(mock_call, retry_options)

        try:
            my_callable()
            self.fail('Should not have been reached')
        except errors.RetryError as exc:
            self.assertIsInstance(exc.cause, CustomException)

        self.assertGreaterEqual(mock_time(),
                                params.total_timeout_millis / _MILLIS_PER_SEC)

        # Very rough bounds
        calls_lower_bound = params.total_timeout_millis / (
            params.max_retry_delay_millis + params.max_rpc_timeout_millis)
        self.assertGreater(mock_call.call_count, calls_lower_bound)

        calls_upper_bound = (params.total_timeout_millis /
                             params.initial_retry_delay_millis)
        self.assertLess(mock_call.call_count, calls_upper_bound)



# Copyright 2017, Google Inc. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# EDITING INSTRUCTIONS
# This file was generated from the file
# https://github.com/google/googleapis/blob/master/google/longrunning/operations.proto,
# and updates to that file get reflected here through a refresh process.
# For the short term, the refresh process will only be runnable by Google engineers.
#
# The only allowed edits are to method and file documentation. A 3-way
# merge preserves those additions if the generated source changes.
"""Accesses the google.longrunning Operations API."""

import collections
import json
import os
import pkg_resources
import platform

from google.gax import api_callable
from google.gax import config
from google.gax import path_template
import google.gax

from google.longrunning import operations_pb2

_PageDesc = google.gax.PageDescriptor


class OperationsClient(object):
    """
    Manages long-running operations with an API service.

    When an API method normally takes long time to complete, it can be designed
    to return ``Operation`` to the client, and the client can use this
    interface to receive the real response asynchronously by polling the
    operation resource, or pass the operation resource to another API (such as
    Google Cloud Pub/Sub API) to receive the response.  Any API service that
    returns long-running operations should implement the ``Operations`` interface
    so developers can have a consistent client experience.
    """

    SERVICE_ADDRESS = 'longrunning.googleapis.com'
    """The default address of the service."""

    DEFAULT_SERVICE_PORT = 443
    """The default port of the service."""

    _PAGE_DESCRIPTORS = {
        'list_operations': _PageDesc('page_token', 'next_page_token',
                                     'operations')
    }

    # The scopes needed to make gRPC calls to all of the methods defined in
    # this service
    _ALL_SCOPES = ()

    def __init__(self,
                 service_path=SERVICE_ADDRESS,
                 port=DEFAULT_SERVICE_PORT,
                 channel=None,
                 credentials=None,
                 ssl_credentials=None,
                 scopes=None,
                 client_config=None,
                 app_name=None,
                 app_version='',
                 lib_name=None,
                 lib_version='',
                 metrics_headers=()):
        """Constructor.

        Args:
          service_path (string): The domain name of the API remote host.
          port (int): The port on which to connect to the remote host.
          channel (:class:`grpc.Channel`): A ``Channel`` instance through
            which to make calls.
          credentials (object): The authorization credentials to attach to
            requests. These credentials identify this application to the
            service.
          ssl_credentials (:class:`grpc.ChannelCredentials`): A
            ``ChannelCredentials`` instance for use with an SSL-enabled
            channel.
          scopes (list[string]): A list of OAuth2 scopes to attach to requests.
          client_config (dict):
            A dictionary for call options for each method. See
            :func:`google.gax.construct_settings` for the structure of
            this data. Falls back to the default config if not specified
            or the specified config is missing data points.
          app_name (string): The name of the application calling
            the service. Recommended for analytics purposes.
          app_version (string): The version of the application calling
            the service. Recommended for analytics purposes.
          lib_name (string): The API library software used for calling
            the service. (Unless you are writing an API client itself,
            leave this as default.)
          lib_version (string): The API library software version used
            for calling the service. (Unless you are writing an API client
            itself, leave this as default.)
          metrics_headers (dict): A dictionary of values for tracking
            client library metrics. Ultimately serializes to a string
            (e.g. 'foo/1.2.3 bar/3.14.1'). This argument should be
            considered private.

        Returns:
          A OperationsClient object.
        """
        # Unless the calling application specifically requested
        # OAuth scopes, request everything.
        if scopes is None:
            scopes = self._ALL_SCOPES

        # Initialize an empty client config, if none is set.
        if client_config is None:
            client_config = {}

        # Initialize metrics_headers as an ordered dictionary
        # (cuts down on cardinality of the resulting string slightly).
        metrics_headers = collections.OrderedDict(metrics_headers)
        metrics_headers['gl-python'] = platform.python_version()

        # The library may or may not be set, depending on what is
        # calling this client. Newer client libraries set the library name
        # and version.
        if lib_name:
            metrics_headers[lib_name] = lib_version

        # Load the configuration defaults.
        default_client_config = json.loads(
            pkg_resources.resource_string(
                __name__, 'operations_client_config.json').decode())
        defaults = api_callable.construct_settings(
            'google.longrunning.Operations',
            default_client_config,
            client_config,
            config.STATUS_CODE_NAMES,
            metrics_headers=metrics_headers,
            page_descriptors=self._PAGE_DESCRIPTORS, )
        self.operations_stub = config.create_stub(
            operations_pb2.OperationsStub,
            channel=channel,
            service_path=service_path,
            service_port=port,
            credentials=credentials,
            scopes=scopes,
            ssl_credentials=ssl_credentials)

        self._get_operation = api_callable.create_api_call(
            self.operations_stub.GetOperation,
            settings=defaults['get_operation'])
        self._list_operations = api_callable.create_api_call(
            self.operations_stub.ListOperations,
            settings=defaults['list_operations'])
        self._cancel_operation = api_callable.create_api_call(
            self.operations_stub.CancelOperation,
            settings=defaults['cancel_operation'])
        self._delete_operation = api_callable.create_api_call(
            self.operations_stub.DeleteOperation,
            settings=defaults['delete_operation'])

    # Service calls
    def get_operation(self, name, options=None):
        """
        Gets the latest state of a long-running operation.  Clients can use this
        method to poll the operation result at intervals as recommended by the API
        service.

        Example:
          >>> from google.gapic.longrunning import operations_client
          >>> api = operations_client.OperationsClient()
          >>> name = ''
          >>> response = api.get_operation(name)

        Args:
          name (string): The name of the operation resource.
          options (:class:`google.gax.CallOptions`): Overrides the default
            settings for this call, e.g, timeout, retries etc.

        Returns:
          A :class:`google.longrunning.operations_pb2.Operation` instance.

        Raises:
          :exc:`google.gax.errors.GaxError` if the RPC is aborted.
          :exc:`ValueError` if the parameters are invalid.
        """
        # Create the request object.
        request = operations_pb2.GetOperationRequest(name=name)
        return self._get_operation(request, options)

    def list_operations(self, name, filter_, page_size=0, options=None):
        """
        Lists operations that match the specified filter in the request. If the
        server doesn't support this method, it returns ``UNIMPLEMENTED``.
        NOTE: the ``name`` binding below allows API services to override the binding
        to use different resource name schemes, such as ``users/*/operations``.

        Example:
          >>> from google.gapic.longrunning import operations_client
          >>> from google.gax import CallOptions, INITIAL_PAGE
          >>> api = operations_client.OperationsClient()
          >>> name = ''
          >>> filter_ = ''
          >>>
          >>> # Iterate over all results
          >>> for element in api.list_operations(name, filter_):
          >>>   # process element
          >>>   pass
          >>>
          >>> # Or iterate over results one page at a time
          >>> for page in api.list_operations(name, filter_, options=CallOptions(page_token=INITIAL_PAGE)):
          >>>   for element in page:
          >>>     # process element
          >>>     pass

        Args:
          name (string): The name of the operation collection.
          filter_ (string): The standard list filter.
          page_size (int): The maximum number of resources contained in the
            underlying API response. If page streaming is performed per-
            resource, this parameter does not affect the return value. If page
            streaming is performed per-page, this determines the maximum number
            of resources in a page.
          options (:class:`google.gax.CallOptions`): Overrides the default
            settings for this call, e.g, timeout, retries etc.

        Returns:
          A :class:`google.gax.PageIterator` instance. By default, this
          is an iterable of :class:`google.longrunning.operations_pb2.Operation` instances.
          This object can also be configured to iterate over the pages
          of the response through the `CallOptions` parameter.

        Raises:
          :exc:`google.gax.errors.GaxError` if the RPC is aborted.
          :exc:`ValueError` if the parameters are invalid.
        """
        # Create the request object.
        request = operations_pb2.ListOperationsRequest(
            name=name, filter=filter_, page_size=page_size)
        return self._list_operations(request, options)

    def cancel_operation(self, name, options=None):
        """
        Starts asynchronous cancellation on a long-running operation.  The server
        makes a best effort to cancel the operation, but success is not
        guaranteed.  If the server doesn't support this method, it returns
        ``google.rpc.Code.UNIMPLEMENTED``.  Clients can use
        ``Operations.GetOperation`` or
        other methods to check whether the cancellation succeeded or whether the
        operation completed despite cancellation. On successful cancellation,
        the operation is not deleted; instead, it becomes an operation with
        an ``Operation.error`` value with a ``google.rpc.Status.code`` of 1,
        corresponding to ``Code.CANCELLED``.

        Example:
          >>> from google.gapic.longrunning import operations_client
          >>> api = operations_client.OperationsClient()
          >>> name = ''
          >>> api.cancel_operation(name)

        Args:
          name (string): The name of the operation resource to be cancelled.
          options (:class:`google.gax.CallOptions`): Overrides the default
            settings for this call, e.g, timeout, retries etc.

        Raises:
          :exc:`google.gax.errors.GaxError` if the RPC is aborted.
          :exc:`ValueError` if the parameters are invalid.
        """
        # Create the request object.
        request = operations_pb2.CancelOperationRequest(name=name)
        self._cancel_operation(request, options)

    def delete_operation(self, name, options=None):
        """
        Deletes a long-running operation. This method indicates that the client is
        no longer interested in the operation result. It does not cancel the
        operation. If the server doesn't support this method, it returns
        ``google.rpc.Code.UNIMPLEMENTED``.

        Example:
          >>> from google.gapic.longrunning import operations_client
          >>> api = operations_client.OperationsClient()
          >>> name = ''
          >>> api.delete_operation(name)

        Args:
          name (string): The name of the operation resource to be deleted.
          options (:class:`google.gax.CallOptions`): Overrides the default
            settings for this call, e.g, timeout, retries etc.

        Raises:
          :exc:`google.gax.errors.GaxError` if the RPC is aborted.
          :exc:`ValueError` if the parameters are invalid.
        """
        # Create the request object.
        request = operations_pb2.DeleteOperationRequest(name=name)
        self._delete_operation(request, options)



# Copyright 2016, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pylint: disable=missing-docstring,no-self-use,no-init,invalid-name
# pylint: disable=protected-access
"""Unit tests for _grpc_google_auth."""

from __future__ import absolute_import

import sys

import google_auth_httplib2
import mock
from six.moves import reload_module
import unittest2

import google.auth.credentials
import google.auth.transport.requests
from google.gax import _grpc_google_auth


class TestGetDefaultCredentials(unittest2.TestCase):
    @mock.patch('google.auth.default', autospec=True)
    def test(self, default_mock):
        default_mock.return_value = (
            mock.sentinel.credentials, mock.sentinel.project)
        scopes = ['fake', 'scopes']

        got = _grpc_google_auth.get_default_credentials(scopes)

        self.assertEqual(got, mock.sentinel.credentials)
        default_mock.assert_called_once_with(scopes)


class TestSecureAuthorizedChannel(unittest2.TestCase):
    @mock.patch('google.gax._grpc_google_auth._request_factory', autospec=True)
    @mock.patch(
        'google.auth.transport.grpc.secure_authorized_channel', autospec=True)
    def test(self, secure_authorized_channel_mock, request_factory_mock):

        got_channel = _grpc_google_auth.secure_authorized_channel(
            mock.sentinel.credentials,
            mock.sentinel.target)

        request_factory_mock.assert_called_once_with()
        secure_authorized_channel_mock.assert_called_once_with(
            mock.sentinel.credentials,
            request_factory_mock.return_value,
            mock.sentinel.target,
            ssl_credentials=None)

        self.assertEqual(
            got_channel, secure_authorized_channel_mock.return_value)


class TestRequestsRequestFactory(unittest2.TestCase):
    def test(self):
        self.assertEqual(
            _grpc_google_auth._request_factory,
            google.auth.transport.requests.Request)


class TestHttplib2RequestFactory(unittest2.TestCase):
    def setUp(self):
        # Block requests module during this test.
        self._requests_module = sys.modules['google.auth.transport.requests']
        sys.modules['google.auth.transport.requests'] = None
        reload_module(_grpc_google_auth)

    def tearDown(self):
        sys.modules['google.auth.transport.requests'] = self._requests_module
        reload_module(_grpc_google_auth)

    def test(self):
        request = _grpc_google_auth._request_factory()
        self.assertIsInstance(request, google_auth_httplib2.Request)


class TestNoHttpClient(unittest2.TestCase):
    def setUp(self):
        # Block all transport modules during this test.
        self._requests_module = sys.modules['google.auth.transport.requests']
        self._httplib2_module = sys.modules['google_auth_httplib2']
        sys.modules['google.auth.transport.requests'] = None
        sys.modules['google_auth_httplib2'] = None

    def tearDown(self):
        sys.modules['google.auth.transport.requests'] = self._requests_module
        sys.modules['google_auth_httplib2'] = self._httplib2_module
        reload_module(_grpc_google_auth)

    def test(self):
        with self.assertRaises(ImportError):
            reload_module(_grpc_google_auth)



# Copyright 2016, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pylint: disable=missing-docstring,no-self-use,no-init,invalid-name
"""Unit tests for grpc."""

from __future__ import absolute_import

import mock
import unittest2

from google.gax import grpc


def _fake_create_stub(channel):
    return channel


class TestCreateStub(unittest2.TestCase):
    FAKE_SERVICE_PATH = 'service_path'
    FAKE_PORT = 10101

    @mock.patch('google.gax._grpc_google_auth.get_default_credentials')
    @mock.patch('google.gax._grpc_google_auth.secure_authorized_channel')
    def test_creates_a_stub_with_default_credentials(
            self, secure_authorized_channel, get_default_credentials):
        fake_scopes = ['one', 'two']
        got_channel = grpc.create_stub(
            _fake_create_stub, service_path=self.FAKE_SERVICE_PATH,
            service_port=self.FAKE_PORT, scopes=fake_scopes)

        get_default_credentials.assert_called_once_with(fake_scopes)
        secure_authorized_channel.assert_called_once_with(
            get_default_credentials.return_value,
            '{}:{}'.format(self.FAKE_SERVICE_PATH, self.FAKE_PORT),
            ssl_credentials=None)

        self.assertEqual(got_channel, secure_authorized_channel.return_value)

    @mock.patch('google.gax._grpc_google_auth.get_default_credentials')
    @mock.patch('google.gax._grpc_google_auth.secure_authorized_channel')
    def test_creates_a_stub_with_explicit_credentials(
            self, secure_authorized_channel, get_default_credentials):
        credentials = mock.Mock()
        got_channel = grpc.create_stub(
            _fake_create_stub, service_path=self.FAKE_SERVICE_PATH,
            service_port=self.FAKE_PORT, credentials=credentials)

        self.assertFalse(get_default_credentials.called)
        secure_authorized_channel.assert_called_once_with(
            credentials,
            '{}:{}'.format(self.FAKE_SERVICE_PATH, self.FAKE_PORT),
            ssl_credentials=None)

        self.assertEqual(got_channel, secure_authorized_channel.return_value)

    @mock.patch('google.gax._grpc_google_auth.get_default_credentials')
    @mock.patch('google.gax._grpc_google_auth.secure_authorized_channel')
    def test_creates_a_stub_with_given_channel(
            self, secure_authorized_channel, get_default_credentials):
        fake_channel = mock.Mock()
        got_channel = grpc.create_stub(
            _fake_create_stub, channel=fake_channel)
        self.assertEqual(got_channel, fake_channel)
        self.assertFalse(secure_authorized_channel.called)
        self.assertFalse(get_default_credentials.called)


class TestErrors(unittest2.TestCase):
    class MyError(grpc.RpcError):
        def code(self):
            return grpc.StatusCode.UNKNOWN

    def test_exc_to_code(self):
        code = grpc.exc_to_code(TestErrors.MyError())
        self.assertEqual(code, grpc.StatusCode.UNKNOWN)
        self.assertEqual(code, grpc.STATUS_CODE_NAMES['UNKNOWN'])
        self.assertIsNone(grpc.exc_to_code(Exception))
        self.assertIsNone(grpc.exc_to_code(grpc.RpcError()))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import io

from setuptools import find_packages
from setuptools import setup


DEPENDENCIES = [
    'dill >= 0.2.5, < 0.3dev',
    'future >= 0.16.0, < 0.17dev',
    'googleapis-common-protos >= 1.5.2, < 2.0dev',
    'grpcio >=1.0.2, < 2.0dev',
    'google-auth >= 1.0.0, <2.0dev',
    'ply == 3.8',
    'protobuf >= 3.0.0, < 4.0dev',
    'requests >= 2.13.0, < 3.0dev'
]

with io.open('README.rst', 'r') as readme:
    long_description = readme.read()

setup(
    name='google-gax',
    version='0.15.11',
    description='Google API Extensions',
    long_description=long_description,
    author='Google API Authors',
    author_email='googleapis-packages@google.com',
    url='https://github.com/googleapis/gax-python',
    packages=find_packages(exclude=('tests*',)),
    namespace_packages=('google',),
    license='BSD-3-Clause',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    install_requires=DEPENDENCIES,
    include_package_data=True,
)
